import asyncio
import contextlib
import logging
from abc import ABC
from datetime import datetime, timedelta, timezone
from typing import Dict, List, Literal, Optional, Tuple, Union, cast

import discord

from redbot.core.bot import Red
from redbot.core import commands, i18n, modlog, Config
from redbot.core.utils import AsyncIter, bounded_gather, can_user_react_in
from redbot.core.utils.chat_formatting import (
    bold,
    humanize_timedelta,
    humanize_list,
    inline,
    pagify,
)
from redbot.core.utils.mod import get_audit_reason
from redbot.core.utils.menus import start_adding_reactions
from redbot.core.utils.views import SimpleMenu
from redbot.core.utils.predicates import MessagePredicate, ReactionPredicate

from .converters import MuteTime
from .models import ChannelMuteResponse, MuteResponse
from .voicemutes import VoiceMutes

T_ = i18n.Translator("Mutes", __file__)

_ = lambda s: s

MUTE_UNMUTE_ISSUES = {
    "already_muted": _("That user is already muted in {location}."),
    "already_unmuted": _("That user is not muted in {location}."),
    "hierarchy_problem": _(
        "I cannot let you do that. You are not higher than the user in the role hierarchy."
    ),
    "assigned_role_hierarchy_problem": _(
        "I cannot let you do that. You are not higher than the mute role in the role hierarchy."
    ),
    "is_admin": _("That user cannot be (un)muted, as they have the Administrator permission."),
    "permissions_issue_role": _(
        "Failed to mute or unmute user. I need the Manage Roles "
        "permission and the user I'm muting must be "
        "lower than myself in the role hierarchy."
    ),
    "permissions_issue_guild": _(
        "Failed to mute or unmute user. I need the Timeout Members "
        "permission and the user I'm muting must be "
        "lower than myself in the role hierarchy."
    ),
    "permissions_issue_channel": _(
        "Failed to mute or unmute user. I need the Manage Permissions permission in {location}."
    ),
    "left_guild": _("The user has left the server while applying an overwrite."),
    "unknown_channel": _("The channel I tried to mute or unmute the user in isn't found."),
    "role_missing": _("The mute role no longer exists."),
    "voice_mute_permission": _(
        "Because I don't have the Move Members permission, this will take into effect when the user rejoins."
    ),
    "mute_is_too_long": _("Timeouts cannot be longer than 28 days."),
    "timeouts_require_time": _("You must provide a time for the timeout to end."),
    "is_not_voice_mute": _(
        "That user is channel muted in their current voice channel, not just voice muted."
        " If you want to fully unmute this user in the channel,"
        " use {command} in their voice channel's text channel instead."
    ),
}
_ = T_

log = logging.getLogger("red.cogs.mutes")

__version__ = "1.0.0"


class CompositeMetaClass(type(commands.Cog), type(ABC)):
    """
    This allows the metaclass used for proper type detection to
    coexist with discord.py's metaclass
    """

    pass


@i18n.cog_i18n(_)
class Mutes(VoiceMutes, commands.Cog, metaclass=CompositeMetaClass):
    """
    Mute users temporarily or indefinitely.
    """

    def __init__(self, bot: Red):
        self.bot = bot
        self.config = Config.get_conf(self, 49615220001, force_registration=True)
        default_guild = {
            "mute_role": None,
            "notification_channel": None,
            "muted_users": {},
            "default_time": 0,
            "dm": False,
            "show_mod": False,
        }
        self.config.register_global(schema_version=0)
        self.config.register_guild(**default_guild)
        self.config.register_member(perms_cache={})
        self.config.register_channel(muted_users={})
        self._server_mutes: Dict[int, Dict[int, dict]] = {}
        self._channel_mutes: Dict[int, Dict[int, dict]] = {}
        self._unmute_tasks: Dict[str, asyncio.Task] = {}
        self._unmute_task: Optional[asyncio.Task] = None
        self.mute_role_cache: Dict[int, int] = {}
        # this is a dict of guild ID's and asyncio.Events
        # to wait for a guild to finish channel unmutes before
        # checking for manual overwrites
        self._channel_mute_events: Dict[int, asyncio.Event] = {}
        self._ready = asyncio.Event()
        self._init_task: Optional[asyncio.Task] = None
        self._ready_raised = False

    def create_init_task(self) -> None:
        def _done_callback(task: asyncio.Task) -> None:
            try:
                exc = task.exception()
            except asyncio.CancelledError:
                pass
            else:
                if exc is None:
                    return
                log.error(
                    "An unexpected error occurred during Mutes's initialization.",
                    exc_info=exc,
                )
            self._ready_raised = True
            self._ready.set()

        self._init_task = asyncio.create_task(self.initialize())
        self._init_task.add_done_callback(_done_callback)

    async def red_delete_data_for_user(
        self,
        *,
        requester: Literal["discord_deleted_user", "owner", "user", "user_strict"],
        user_id: int,
    ):
        """Mutes are considered somewhat critical
        Therefore the only data that we should delete
        is that which comes from discord requesting us to
        remove data about a user
        """
        if requester != "discord_deleted_user":
            return

        await self._ready.wait()
        if self._ready_raised:
            raise RuntimeError(
                "Mutes cog is in a bad state, can't proceed with data deletion request."
            )
        all_members = await self.config.all_members()
        for g_id, data in all_members.items():
            for m_id, mutes in data.items():
                if m_id == user_id:
                    await self.config.member_from_ids(g_id, m_id).clear()

    async def initialize(self):
        await self.bot.wait_until_red_ready()
        await self._maybe_update_config()

        guild_data = await self.config.all_guilds()
        for g_id, mutes in guild_data.items():
            self._server_mutes[g_id] = {}
            if mutes["mute_role"]:
                self.mute_role_cache[g_id] = mutes["mute_role"]
            for user_id, mute in mutes["muted_users"].items():
                self._server_mutes[g_id][int(user_id)] = mute
        channel_data = await self.config.all_channels()
        for c_id, mutes in channel_data.items():
            self._channel_mutes[c_id] = {}
            for user_id, mute in mutes["muted_users"].items():
                self._channel_mutes[c_id][int(user_id)] = mute
        self._unmute_task = asyncio.create_task(self._handle_automatic_unmute())
        self._ready.set()

    async def _maybe_update_config(self):
        schema_version = await self.config.schema_version()

        if schema_version == 0:
            await self._schema_0_to_1()
            schema_version += 1
            await self.config.schema_version.set(schema_version)

    async def _schema_0_to_1(self):
        """This contains conversion that adds guild ID to channel mutes data."""
        all_channels = await self.config.all_channels()
        if not all_channels:
            return

        start = datetime.now()
        log.info(
            "Config conversion to schema_version 1 started. This may take a while to proceed..."
        )
        async for channel_id in AsyncIter(all_channels.keys()):
            try:
                if (channel := self.bot.get_channel(channel_id)) is None:
                    channel = await self.bot.fetch_channel(channel_id)
                async with self.config.channel_from_id(channel_id).muted_users() as muted_users:
                    for mute_id, mute_data in muted_users.items():
                        mute_data["guild"] = channel.guild.id
            except (discord.NotFound, discord.Forbidden):
                await self.config.channel_from_id(channel_id).clear()

        log.info(
            "Config conversion to schema_version 1 done. It took %s to proceed.",
            datetime.now() - start,
        )

    async def cog_before_invoke(self, ctx: commands.Context):
        if not self._ready.is_set():
            async with ctx.typing():
                await self._ready.wait()
        if self._ready_raised:
            await ctx.send(
                "There was an error during Mutes's initialization."
                " Check logs for more information."
            )
            raise commands.CheckFailure()

    def cog_unload(self):
        if self._init_task is not None:
            self._init_task.cancel()
        if self._unmute_task is not None:
            self._unmute_task.cancel()
        for task in self._unmute_tasks.values():
            task.cancel()

    async def is_allowed_by_hierarchy(
        self, guild: discord.Guild, mod: discord.Member, user: discord.Member
    ):
        is_special = mod == guild.owner or await self.bot.is_owner(mod)
        return mod.top_role > user.top_role or is_special

    async def _handle_automatic_unmute(self):
        """This is the core task creator and loop
        for automatic unmutes

        A resolution of 30 seconds seems appropriate
        to allow for decent resolution on low timed
        unmutes and without being too busy on our event loop
        """
        await self.bot.wait_until_red_ready()
        await self._ready.wait()
        if self._ready_raised:
            raise RuntimeError("Mutes cog is in a bad state, cancelling automatic unmute task.")
        while True:
            await self._clean_tasks()
            try:
                await self._handle_server_unmutes()
            except Exception:
                log.error("error checking server unmutes", exc_info=True)
            await asyncio.sleep(0.1)
            try:
                await self._handle_channel_unmutes()
            except Exception:
                log.error("error checking channel unmutes", exc_info=True)
            await asyncio.sleep(30)

    async def _clean_tasks(self):
        """This is here to cleanup our tasks
        and log when we have something going wrong
        inside our tasks.
        """
        log.debug("Cleaning unmute tasks")
        for task_id in list(self._unmute_tasks.keys()):
            task = self._unmute_tasks[task_id]

            if task.cancelled():
                self._unmute_tasks.pop(task_id, None)
                continue

            if task.done():
                try:
                    r = task.result()
                except Exception as exc:
                    log.error("An unexpected error occurred in the unmute task", exc_info=exc)
                self._unmute_tasks.pop(task_id, None)

    async def _handle_server_unmutes(self):
        """This is where the logic for role unmutes is taken care of"""
        log.debug("Checking server unmutes")
        for g_id in self._server_mutes:
            guild = self.bot.get_guild(g_id)
            if guild is None or await self.bot.cog_disabled_in_guild(self, guild):
                continue
            await i18n.set_contextual_locales_from_guild(self.bot, guild)
            for u_id in self._server_mutes[guild.id]:
                if self._server_mutes[guild.id][u_id]["until"] is None:
                    continue
                time_to_unmute = (
                    self._server_mutes[guild.id][u_id]["until"]
                    - datetime.now(timezone.utc).timestamp()
                )
                if time_to_unmute < 60.0:
                    task_name = f"server-unmute-{g_id}-{u_id}"
                    if task_name in self._unmute_tasks:
                        continue
                    log.debug(f"Creating task: {task_name}")
                    self._unmute_tasks[task_name] = asyncio.create_task(
                        self._auto_unmute_user(guild, self._server_mutes[guild.id][u_id])
                    )

    async def _auto_unmute_user(self, guild: discord.Guild, data: dict):
        """
        This handles role unmutes automatically

        Since channel overwrite mutes are handled under the separate
        _auto_channel_unmute_user methods here we don't
        need to worry about the dict response for message
        since only role based mutes get added here
        """
        delay = data["until"] - datetime.now(timezone.utc).timestamp()
        if delay < 1:
            delay = 0
        await asyncio.sleep(delay)

        member = guild.get_member(data["member"])
        author = guild.get_member(data["author"])
        if not member:
            async with self.config.guild(guild).muted_users() as muted_users:
                if str(data["member"]) in muted_users:
                    del muted_users[str(data["member"])]
            del self._server_mutes[guild.id][data["member"]]
            return
        result = await self.unmute_user(guild, None, member, _("Automatic unmute"))
        async with self.config.guild(guild).muted_users() as muted_users:
            if str(member.id) in muted_users:
                del muted_users[str(member.id)]
        if result.success:
            await modlog.create_case(
                self.bot,
                guild,
                datetime.now(timezone.utc),
                "sunmute",
                member,
                author,
                _("Automatic unmute"),
                until=None,
            )
            await self._send_dm_notification(
                member, author, guild, _("Server unmute"), _("Automatic unmute")
            )
        else:
            chan_id = await self.config.guild(guild).notification_channel()
            notification_channel = guild.get_channel(chan_id)
            if not notification_channel:
                return
            if not notification_channel.permissions_for(guild.me).send_messages:
                return
            error_msg = _(
                "I am unable to unmute {user} for the following reason:\n{reason}"
            ).format(user=member, reason=result.reason)
            try:
                await notification_channel.send(error_msg)
            except discord.errors.Forbidden:
                log.info(error_msg)
                return

    async def _handle_channel_unmutes(self):
        """This is where the logic for handling channel unmutes is taken care of"""
        log.debug("Checking channel unmutes")
        multiple_mutes = {}
        for c_id, c_data in self._channel_mutes.items():
            for u_id in self._channel_mutes[c_id]:
                if (
                    not self._channel_mutes[c_id][u_id]
                    or not self._channel_mutes[c_id][u_id]["until"]
                ):
                    continue
                guild = self.bot.get_guild(self._channel_mutes[c_id][u_id]["guild"])
                if guild is None or await self.bot.cog_disabled_in_guild(self, guild):
                    continue
                time_to_unmute = (
                    self._channel_mutes[c_id][u_id]["until"]
                    - datetime.now(timezone.utc).timestamp()
                )
                if time_to_unmute < 60.0:
                    if guild not in multiple_mutes:
                        multiple_mutes[guild] = {}
                    if u_id not in multiple_mutes[guild]:
                        multiple_mutes[guild][u_id] = {c_id: self._channel_mutes[c_id][u_id]}
                    else:
                        multiple_mutes[guild][u_id][c_id] = self._channel_mutes[c_id][u_id]

        for guild, users in multiple_mutes.items():
            await i18n.set_contextual_locales_from_guild(self.bot, guild)
            for user, channels in users.items():
                if len(channels) > 1:
                    task_name = f"server-unmute-channels-{guild.id}-{user}"
                    if task_name in self._unmute_tasks:
                        continue
                    log.debug(f"Creating task: {task_name}")
                    member = guild.get_member(user)
                    self._unmute_tasks[task_name] = asyncio.create_task(
                        self._auto_channel_unmute_user_multi(member, guild, channels)
                    )

                else:
                    for channel, mute_data in channels.items():
                        task_name = f"channel-unmute-{channel}-{user}"
                        log.debug(f"Creating task: {task_name}")
                        if task_name in self._unmute_tasks:
                            continue
                        if guild_channel := guild.get_channel(channel):
                            self._unmute_tasks[task_name] = asyncio.create_task(
                                self._auto_channel_unmute_user(guild_channel, mute_data)
                            )

        del multiple_mutes

    async def _auto_channel_unmute_user_multi(
        self, member: discord.Member, guild: discord.Guild, channels: Dict[int, dict]
    ):
        """This is meant to handle multiple channels all being unmuted at once"""
        if guild.id in self._channel_mute_events:
            self._channel_mute_events[guild.id].clear()
        else:
            self._channel_mute_events[guild.id] = asyncio.Event()
        tasks = []
        for channel, mute_data in channels.items():
            author = guild.get_member(mute_data["author"])
            tasks.append(
                self._auto_channel_unmute_user(guild.get_channel(channel), mute_data, False)
            )
        results = await bounded_gather(*tasks)
        unmuted_channels = [guild.get_channel(c) for c in channels.keys()]
        for result in results:
            if not result:
                continue
            _member, channel, reason = result
            unmuted_channels.remove(channel)
        modlog_reason = _("Automatic unmute")

        channel_list = humanize_list([c.mention for c in unmuted_channels if c is not None])
        if channel_list:
            modlog_reason += _("\nUnmuted in channels: ") + channel_list

        await modlog.create_case(
            self.bot,
            guild,
            datetime.now(timezone.utc),
            "sunmute",
            member,
            author,
            modlog_reason,
            until=None,
        )
        await self._send_dm_notification(
            member, author, guild, _("Server unmute"), _("Automatic unmute")
        )
        self._channel_mute_events[guild.id].set()
        if any(results):
            reasons = {}
            for result in results:
                if not result:
                    continue
                _member, channel, reason = result
                if reason not in reasons:
                    reasons[reason] = [channel]
                else:
                    reasons[reason].append(channel)
            error_msg = _("{member} could not be unmuted for the following reasons:\n").format(
                member=member
            )
            for reason, channel_list in reasons.items():
                error_msg += _("{reason} In the following channels: {channels}\n").format(
                    reason=reason,
                    channels=humanize_list([c.mention for c in channel_list]),
                )
            chan_id = await self.config.guild(guild).notification_channel()
            notification_channel = guild.get_channel(chan_id)
            if notification_channel is None:
                return None
            if not notification_channel.permissions_for(guild.me).send_messages:
                return None
            try:
                await notification_channel.send(error_msg)
            except discord.errors.Forbidden:
                log.info(error_msg)
                return None

    async def _auto_channel_unmute_user(
        self, channel: discord.abc.GuildChannel, data: dict, create_case: bool = True
    ) -> Optional[Tuple[discord.Member, discord.abc.GuildChannel, str]]:
        """This is meant to unmute a user in individual channels"""
        delay = data["until"] - datetime.now(timezone.utc).timestamp()
        if delay < 1:
            delay = 0
        await asyncio.sleep(delay)
        member = channel.guild.get_member(data["member"])
        author = channel.guild.get_member(data["author"])
        if not member:
            async with self.config.channel(channel).muted_users() as muted_users:
                if str(data["member"]) in muted_users:
                    del muted_users[str(data["member"])]
            if (
                channel.id in self._channel_mutes
                and data["member"] in self._channel_mutes[channel.id]
            ):
                del self._channel_mutes[channel.id][data["member"]]
            return None
        result = await self.channel_unmute_user(
            channel.guild, channel, None, member, _("Automatic unmute")
        )
        async with self.config.channel(channel).muted_users() as muted_users:
            if str(member.id) in muted_users:
                del muted_users[str(member.id)]
        if result.success:
            if create_case:
                if data.get("voice_mute", False):
                    unmute_type = "vunmute"
                    notification_title = _("Voice unmute")
                else:
                    unmute_type = "cunmute"
                    notification_title = _("Channel unmute")
                await modlog.create_case(
                    self.bot,
                    channel.guild,
                    datetime.now(timezone.utc),
                    unmute_type,
                    member,
                    channel.guild.me,
                    _("Automatic unmute"),
                    until=None,
                    channel=channel,
                )
                await self._send_dm_notification(
                    member, author, channel.guild, notification_title, _("Automatic unmute")
                )
            return None
        else:
            error_msg = _(
                "I am unable to unmute {user} in {channel} for the following reason:\n{reason}"
            ).format(user=member, channel=channel.mention, reason=result.reason)
            if create_case:
                chan_id = await self.config.guild(channel.guild).notification_channel()
                notification_channel = channel.guild.get_channel(chan_id)
                if not notification_channel:
                    return None
                if not notification_channel.permissions_for(channel.guild.me).send_messages:
                    return None
                try:
                    await notification_channel.send(error_msg)
                except discord.errors.Forbidden:
                    log.info(error_msg)
                    return None
            else:
                return (member, channel, result.reason)

    async def _send_dm_notification(
        self,
        user: Union[discord.User, discord.Member],
        moderator: Optional[Union[discord.User, discord.Member]],
        guild: discord.Guild,
        mute_type: str,
        reason: Optional[str],
        duration=None,
    ):
        if user.bot:
            return

        if not await self.config.guild(guild).dm():
            return

        show_mod = await self.config.guild(guild).show_mod()
        title = bold(mute_type)
        if duration:
            duration_str = humanize_timedelta(timedelta=duration)
            until = datetime.now(timezone.utc) + duration
            until_str = discord.utils.format_dt(until)

        if moderator is None:
            moderator_str = _("Unknown")
        else:
            moderator_str = str(moderator)

        if not reason:
            reason = _("No reason provided.")

        if await self.bot.embed_requested(user):
            em = discord.Embed(
                title=title,
                description=reason,
                color=await self.bot.get_embed_color(user),
            )
            em.timestamp = datetime.now(timezone.utc)
            if duration:
                em.add_field(name=_("Until"), value=until_str)
                em.add_field(name=_("Duration"), value=duration_str)
            em.add_field(name=_("Guild"), value=guild.name, inline=False)
            if show_mod:
                em.add_field(name=_("Moderator"), value=moderator_str)
            try:
                await user.send(embed=em)
            except discord.Forbidden:
                pass
        else:
            message = f"{title}\n>>> "
            message += reason
            message += (f"\n{bold(_('Moderator:'))} {moderator_str}") if show_mod else ""
            message += (
                (f"\n{bold(_('Until:'))} {until_str}\n{bold(_('Duration:'))} {duration_str}")
                if duration
                else ""
            )
            message += f"\n{bold(_('Guild:'))} {guild.name}"
            try:
                await user.send(message)
            except discord.Forbidden:
                pass

    @commands.Cog.listener()
    async def on_member_update(self, before: discord.Member, after: discord.Member):
        """
        Used to handle the cache if a member manually has the muted role removed
        """
        guild = before.guild
        if await self.bot.cog_disabled_in_guild(self, guild):
            return
        if guild.id not in self.mute_role_cache:
            return
        should_save = False
        mute_role_id = self.mute_role_cache[guild.id]
        mute_role = guild.get_role(mute_role_id)
        if not mute_role:
            return
        b = set(before.roles)
        a = set(after.roles)
        roles_removed = list(b - a)
        roles_added = list(a - b)
        await i18n.set_contextual_locales_from_guild(self.bot, guild)
        if mute_role in roles_removed:
            # send modlog case for unmute and remove from cache
            if guild.id not in self._server_mutes:
                # they weren't a tracked mute so we can return early
                return
            if after.id in self._server_mutes[guild.id]:
                await modlog.create_case(
                    self.bot,
                    guild,
                    datetime.now(timezone.utc),
                    "sunmute",
                    after,
                    None,
                    _("Manually removed mute role"),
                )
                del self._server_mutes[guild.id][after.id]
                should_save = True
                await self._send_dm_notification(
                    after, None, guild, _("Server unmute"), _("Manually removed mute role")
                )
        elif mute_role in roles_added:
            # send modlog case for mute and add to cache
            if guild.id not in self._server_mutes:
                # initialize the guild in the cache to prevent keyerrors
                self._server_mutes[guild.id] = {}
            if after.id not in self._server_mutes[guild.id]:
                await modlog.create_case(
                    self.bot,
                    guild,
                    datetime.now(timezone.utc),
                    "smute",
                    after,
                    None,
                    _("Manually applied mute role"),
                )
                self._server_mutes[guild.id][after.id] = {
                    "author": None,
                    "member": after.id,
                    "until": None,
                }
                should_save = True
                await self._send_dm_notification(
                    after, None, guild, _("Server mute"), _("Manually applied mute role")
                )
        if should_save:
            await self.config.guild(guild).muted_users.set(self._server_mutes[guild.id])

    @commands.Cog.listener()
    async def on_guild_channel_update(
        self, before: discord.abc.GuildChannel, after: discord.abc.GuildChannel
    ):
        """
        This handles manually removing overwrites for a user that has been muted
        """
        if await self.bot.cog_disabled_in_guild(self, after.guild):
            return
        await i18n.set_contextual_locales_from_guild(self.bot, after.guild)
        if after.guild.id in self._channel_mute_events:
            await self._channel_mute_events[after.guild.id].wait()
        if after.id in self._channel_mutes:
            before_perms: Dict[int, Dict[str, Optional[bool]]] = {
                o.id: {name: attr for name, attr in p} for o, p in before.overwrites.items()
            }
            after_perms: Dict[int, Dict[str, Optional[bool]]] = {
                o.id: {name: attr for name, attr in p} for o, p in after.overwrites.items()
            }
            to_del: List[int] = []
            for user_id, mute_data in self._channel_mutes[after.id].items():
                unmuted = False
                voice_mute = mute_data.get("voice_mute", False)
                if user_id in after_perms:
                    perms_to_check = ["speak"]
                    if not voice_mute:
                        perms_to_check.extend(
                            (
                                "send_messages",
                                "send_messages_in_threads",
                                "create_public_threads",
                                "create_private_threads",
                            )
                        )
                    for perm_name in perms_to_check:
                        unmuted = unmuted or after_perms[user_id][perm_name] is not False
                # explicit is better than implicit :thinkies:
                if user_id in before_perms and (user_id not in after_perms or unmuted):
                    user = after.guild.get_member(user_id)
                    send_dm_notification = True
                    if not user:
                        send_dm_notification = False
                        user = discord.Object(id=user_id)
                    log.debug(f"{user} - {type(user)}")
                    to_del.append(user_id)
                    log.debug("creating case")
                    if voice_mute:
                        unmute_type = "vunmute"
                        notification_title = _("Voice unmute")
                    else:
                        unmute_type = "cunmute"
                        notification_title = _("Channel unmute")
                    if send_dm_notification:
                        await self._send_dm_notification(
                            user,
                            None,
                            after.guild,
                            notification_title,
                            _("Manually removed channel overwrites"),
                        )
                    await modlog.create_case(
                        self.bot,
                        after.guild,
                        datetime.now(timezone.utc),
                        unmute_type,
                        user,
                        None,
                        _("Manually removed channel overwrites"),
                        until=None,
                        channel=after,
                    )
                    log.debug("created case")
            if to_del:
                for u_id in to_del:
                    del self._channel_mutes[after.id][u_id]
                await self.config.channel(after).muted_users.set(self._channel_mutes[after.id])

    @commands.Cog.listener()
    async def on_member_join(self, member: discord.Member):
        guild = member.guild
        if await self.bot.cog_disabled_in_guild(self, guild):
            return
        mute_role = await self.config.guild(guild).mute_role()
        if not mute_role:
            # timeouts already restore on rejoin
            return
        await i18n.set_contextual_locales_from_guild(self.bot, guild)
        if guild.id in self._server_mutes:
            if member.id in self._server_mutes[guild.id]:
                role = guild.get_role(mute_role)
                if not role:
                    return
                if self._server_mutes[guild.id][member.id]["until"]:
                    until = datetime.fromtimestamp(
                        self._server_mutes[guild.id][member.id]["until"]
                    )
                else:
                    until = None
                await self.mute_user(
                    guild, guild.me, member, until, _("Previously muted in this server.")
                )

    @commands.group()
    @commands.guild_only()
    async def muteset(self, ctx: commands.Context):
        """Mute settings."""
        pass

    @muteset.command()
    @commands.guild_only()
    @commands.mod_or_permissions(manage_channels=True)
    async def senddm(self, ctx: commands.Context, true_or_false: bool):
        """Set whether mute notifications should be sent to users in DMs."""
        await self.config.guild(ctx.guild).dm.set(true_or_false)
        if true_or_false:
            await ctx.send(_("I will now try to send mute notifications to users DMs."))
        else:
            await ctx.send(_("Mute notifications will no longer be sent to users DMs."))

    @muteset.command()
    @commands.guild_only()
    @commands.mod_or_permissions(manage_channels=True)
    async def showmoderator(self, ctx, true_or_false: bool):
        """Decide whether the name of the moderator muting a user should be included in the DM to that user."""
        await self.config.guild(ctx.guild).show_mod.set(true_or_false)
        if true_or_false:
            await ctx.send(
                _(
                    "I will include the name of the moderator who issued the mute when sending a DM to a user."
                )
            )
        else:
            await ctx.send(
                _(
                    "I will not include the name of the moderator who issued the mute when sending a DM to a user."
                )
            )

    @muteset.command(name="settings", aliases=["showsettings"])
    @commands.mod_or_permissions(manage_channels=True)
    async def show_mutes_settings(self, ctx: commands.Context):
        """
        Shows the current mute settings for this guild.
        """
        data = await self.config.guild(ctx.guild).all()

        mute_role = ctx.guild.get_role(data["mute_role"])
        notification_channel = ctx.guild.get_channel(data["notification_channel"])
        default_time = timedelta(seconds=data["default_time"])
        msg = _(
            "Mute Role: {role}\n"
            "Notification Channel: {channel}\n"
            "Default Time: {time}\n"
            "Send DM: {dm}\n"
            "Show moderator: {show_mod}"
        ).format(
            role=mute_role.mention if mute_role else _("None"),
            channel=notification_channel.mention if notification_channel else _("None"),
            time=humanize_timedelta(timedelta=default_time) if default_time else _("None"),
            dm=data["dm"],
            show_mod=data["show_mod"],
        )
        await ctx.maybe_send_embed(msg)

    @muteset.command(name="notification")
    @commands.admin_or_permissions(manage_channels=True)
    async def notification_channel_set(
        self,
        ctx: commands.Context,
        channel: Optional[
            Union[discord.TextChannel, discord.VoiceChannel, discord.StageChannel]
        ] = None,
    ):
        """
        Set the notification channel for automatic unmute issues.

        If no channel is provided this will be cleared and notifications
        about issues when unmuting users will not be sent anywhere.
        """
        if channel is None:
            await self.config.guild(ctx.guild).notification_channel.clear()
            await ctx.send(_("Notification channel for unmute issues has been cleared."))
        else:
            await self.config.guild(ctx.guild).notification_channel.set(channel.id)
            await ctx.send(
                _("I will post unmute issues in {channel}.").format(channel=channel.mention)
            )

    @muteset.command(name="role")
    @commands.admin_or_permissions(manage_roles=True)
    @commands.bot_has_guild_permissions(manage_roles=True)
    async def mute_role(self, ctx: commands.Context, *, role: discord.Role = None):
        """Sets the role to be applied when muting a user.

        If no role is setup the bot will attempt to mute a user
        by utilizing server timeouts.

        Note: If no role is setup a user may be able to leave the server
        and rejoin no longer being muted.
        """
        if not role:
            await self.config.guild(ctx.guild).mute_role.set(None)
            if ctx.guild.id in self.mute_role_cache:
                del self.mute_role_cache[ctx.guild.id]
            await ctx.send(_("Discord Timeouts will be used for mutes instead."))
        else:
            if role >= ctx.author.top_role:
                await ctx.send(
                    _("You can't set this role as it is not lower than you in the role hierarchy.")
                )
                return
            await self.config.guild(ctx.guild).mute_role.set(role.id)
            self.mute_role_cache[ctx.guild.id] = role.id
            await ctx.send(_("Mute role set to {role}").format(role=role.name))
        if not await self.config.guild(ctx.guild).notification_channel():
            command_1 = f"`{ctx.clean_prefix}muteset notification`"
            await ctx.send(
                _(
                    "No notification channel has been setup, "
                    "use {command_1} to be updated when there's an issue in automatic unmutes."
                ).format(command_1=command_1)
            )

    @muteset.command(name="makerole")
    @commands.admin_or_permissions(manage_roles=True)
    @commands.bot_has_guild_permissions(manage_roles=True)
    @commands.max_concurrency(1, commands.BucketType.guild)
    async def make_mute_role(self, ctx: commands.Context, *, name: str):
        """Create a Muted role.

        This will create a role and apply overwrites to all available channels
        to more easily setup muting a user.

        If you already have a muted role created on the server use
        `[p]muteset role ROLE_NAME_HERE`
        """
        if await self.config.guild(ctx.guild).mute_role():
            command = f"`{ctx.clean_prefix}muteset role`"
            return await ctx.send(
                _(
                    "There is already a mute role setup in this server. "
                    "Please remove it with {command} before trying to "
                    "create a new one."
                ).format(command=command)
            )
        async with ctx.typing():
            perms = discord.Permissions()
            perms.update(
                send_messages=False,
                send_messages_in_threads=False,
                create_public_threads=False,
                create_private_threads=False,
                use_application_commands=False,
                speak=False,
                add_reactions=False,
            )
            try:
                role = await ctx.guild.create_role(
                    name=name, permissions=perms, reason=_("Mute role setup")
                )
                await self.config.guild(ctx.guild).mute_role.set(role.id)
                # save the role early incase of issue later
            except discord.errors.Forbidden:
                return await ctx.send(_("I could not create a muted role in this server."))
            errors = []
            tasks = []
            for channel in ctx.guild.channels:
                tasks.append(self._set_mute_role_overwrites(role, channel))
            errors = await bounded_gather(*tasks)
            if any(errors):
                msg = _(
                    "I could not set overwrites for the following channels: {channels}"
                ).format(channels=humanize_list([i for i in errors if i]))
                for page in pagify(msg, delims=[" "]):
                    await ctx.send(page)

            await ctx.send(_("Mute role set to {role}").format(role=role.name))
        if not await self.config.guild(ctx.guild).notification_channel():
            command_1 = f"`{ctx.clean_prefix}muteset notification`"
            await ctx.send(
                _(
                    "No notification channel has been setup, "
                    "use {command_1} to be updated when there's an issue in automatic unmutes."
                ).format(command_1=command_1)
            )

    async def _set_mute_role_overwrites(
        self, role: discord.Role, channel: discord.abc.GuildChannel
    ) -> Optional[str]:
        """
        This sets the supplied role and channel overwrites to what we want
        by default for a mute role
        """
        if not channel.permissions_for(channel.guild.me).manage_permissions:
            return channel.mention
        overs = discord.PermissionOverwrite()
        overs.send_messages = False
        overs.send_messages_in_threads = False
        overs.create_public_threads = False
        overs.create_private_threads = False
        overs.use_application_commands = False
        overs.add_reactions = False
        overs.speak = False
        try:
            await channel.set_permissions(role, overwrite=overs, reason=_("Mute role setup"))
            return None
        except discord.errors.Forbidden:
            return channel.mention

    @muteset.command(name="defaulttime", aliases=["time"])
    @commands.mod_or_permissions(manage_messages=True)
    async def default_mute_time(self, ctx: commands.Context, *, time: Optional[MuteTime] = None):
        """
        Set the default mute time for the mute command.

        If no time interval is provided this will be cleared.
        """

        if not time:
            await self.config.guild(ctx.guild).default_time.clear()
            await ctx.send(_("Default mute time removed."))
        else:
            duration = time.get("duration", None)
            if not duration:
                return await ctx.send(_("Please provide a valid time format."))
            if duration >= timedelta(days=365000):
                # prevent setting a default time now that might eventually cause an overflow
                # later as the date goes up. 1000 years gives us approximately 8000 more years
                # of wiggle room.
                return await ctx.send(
                    _("The time provided is too long; use a more reasonable time.")
                )
            await self.config.guild(ctx.guild).default_time.set(duration.total_seconds())
            await ctx.send(
                _("Default mute time set to {time}.").format(
                    time=humanize_timedelta(timedelta=duration)
                )
            )

    async def _check_for_mute_role(self, ctx: commands.Context) -> bool:
        """
        This explains to the user whether or not mutes are setup correctly for
        automatic unmutes.
        """
        command_1 = f"{ctx.clean_prefix}muteset role"
        command_2 = f"{ctx.clean_prefix}muteset makerole"
        msg = _(
            "This server does not have a mute role setup and I do not have permission to timeout users. "
            " You can setup a mute role with {command_1} or"
            " {command_2} if you just want a basic role created setup.\n\n"
        ).format(
            command_1=inline(command_1),
            command_2=inline(command_2),
        )
        mute_role_id = await self.config.guild(ctx.guild).mute_role()
        mute_role = ctx.guild.get_role(mute_role_id)
        timeout_perms = ctx.channel.permissions_for(ctx.me).moderate_members
        if not timeout_perms and not mute_role:
            await ctx.send(msg)
            return False

        return True

    @commands.command()
    @commands.guild_only()
    @commands.mod_or_permissions(manage_roles=True)
    async def activemutes(self, ctx: commands.Context):
        """
        Displays active mutes on this server.
        """
        msg = ""
        if ctx.guild.id in self._server_mutes:
            mutes_data = self._server_mutes[ctx.guild.id]
            if mutes_data:
                msg += _("__Server Mutes__\n")
                for user_id, mutes in mutes_data.items():
                    if not mutes:
                        continue
                    user = ctx.guild.get_member(user_id)
                    if not user:
                        user_str = f"<@!{user_id}>"
                    else:
                        user_str = user.mention
                    if mutes["until"]:
                        timestamp = int(mutes["until"])
                        time_str = discord.utils.format_dt(datetime.fromtimestamp(timestamp))
                    else:
                        time_str = ""
                    msg += f"{user_str} "
                    if time_str:
                        msg += _("__Until__: {time_left}\n").format(time_left=time_str)
                    else:
                        msg += "\n"
        added_timeouts = False
        for member in ctx.guild.members:
            if member.is_timed_out():
                if not added_timeouts:
                    msg += _("__Server Timeouts__\n")
                    added_timeouts = True
                msg += f"{member.mention}"
                time_str = discord.utils.format_dt(member.timed_out_until)
                msg += _("__Until__: {time_left}\n").format(time_left=time_str)
        for channel_id, mutes_data in self._channel_mutes.items():
            if not mutes_data:
                continue
            if ctx.guild.get_channel(channel_id):
                msg += _("__<#{channel_id}> Mutes__\n").format(channel_id=channel_id)
                for user_id, mutes in mutes_data.items():
                    if not mutes:
                        continue
                    user = ctx.guild.get_member(user_id)
                    if not user:
                        user_str = f"<@!{user_id}>"
                    else:
                        user_str = user.mention
                    if mutes["until"]:
                        timestamp = int(mutes["until"])
                        time_str = discord.utils.format_dt(datetime.fromtimestamp(timestamp))
                    else:
                        time_str = ""
                    msg += f"{user_str} "
                    if time_str:
                        msg += _("__Until__: {time_left}\n").format(time_left=time_str)
                    else:
                        msg += "\n"

        if msg:
            msgs = []
            for page in pagify(msg):
                if await ctx.embed_requested():
                    msgs.append(discord.Embed(description=page, colour=await ctx.embed_colour()))
                else:
                    msgs.append(page)
            await SimpleMenu(msgs).start(ctx)
            return
        await ctx.maybe_send_embed(_("There are no mutes on this server right now."))

    @commands.command(usage="<users...> [time_and_reason]")
    @commands.guild_only()
    @commands.mod_or_permissions(moderate_members=True)
    @commands.bot_has_permissions(moderate_members=True)
    async def timeout(
        self,
        ctx: commands.Context,
        users: commands.Greedy[discord.Member],
        *,
        time_and_reason: MuteTime = {},
    ):
        """Timeout users.

        `<users...>` is a space separated list of usernames, ID's, or mentions.
        `[time_and_reason]` is the time to timeout for and reason. Time is
        any valid time length such as `30 minutes` or `2 days`. If nothing
        is provided the timeout will use the set default time or give an error if not set.

        Examples:
        `[p]timeout @member1 @member2 spam 5 hours`
        `[p]timeout @member1 3 days`

        """
        if not users:
            return await ctx.send_help()
        if ctx.me in users:
            return await ctx.send(_("You cannot mute me."))
        if ctx.author in users:
            return await ctx.send(_("You cannot mute yourself."))
        until = time_and_reason.get("until", None)
        reason = time_and_reason.get("reason", None)
        time = ""
        duration = None
        if until:
            duration = time_and_reason.get("duration")
            if duration and duration > timedelta(days=28):
                await ctx.send(_(MUTE_UNMUTE_ISSUES["mute_is_too_long"]))
                return
            length = humanize_timedelta(timedelta=duration)
            time = _(" for {length} until {duration}").format(
                length=length, duration=discord.utils.format_dt(until)
            )

        else:
            default_duration = await self.config.guild(ctx.guild).default_time()
            if default_duration:
                duration = timedelta(seconds=default_duration)
                until = ctx.message.created_at + duration
                length = humanize_timedelta(seconds=default_duration)
                time = _(" for {length} until {duration}").format(
                    length=length, duration=discord.utils.format_dt(until)
                )

        success_list = []
        issues_list = []
        for member in users:
            ret = MuteResponse(success=False, reason=None, user=member)
            if member.guild_permissions >= ctx.author.guild_permissions:
                ret.reason = _(MUTE_UNMUTE_ISSUES["hierarchy_problem"])
                issues_list.append(ret)
                continue
            if member.guild_permissions.administrator:
                ret.reason = _(MUTE_UNMUTE_ISSUES["is_admin"])
                issues_list.append(ret)
                continue

            try:
                await member.edit(timed_out_until=until, reason=reason)
                success_list.append(member)
            except Exception:
                pass
        if success_list:
            msg = _("{users} has been timed out in this server{time}.")
            if len(success_list) > 1:
                msg = _("{users} have been timed out in this server{time}.")
            await ctx.send(
                msg.format(users=humanize_list([f"`{u}`" for u in success_list]), time=time)
            )
        else:
            await ctx.send(_("None of the users provided could be muted properly."))
        if issues_list:
            await self.handle_issues(ctx, issues_list)

    @commands.command(usage="<users...> [time_and_reason]")
    @commands.guild_only()
    @commands.mod_or_permissions(manage_roles=True, moderate_members=True)
    async def mute(
        self,
        ctx: commands.Context,
        users: commands.Greedy[discord.Member],
        *,
        time_and_reason: MuteTime = {},
    ):
        """Mute users.

        `<users...>` is a space separated list of usernames, ID's, or mentions.
        `[time_and_reason]` is the time to mute for and reason. Time is
        any valid time length such as `30 minutes` or `2 days`. If nothing
        is provided the mute will use the set default time or indefinite if not set.

        Examples:
        `[p]mute @member1 @member2 spam 5 hours`
        `[p]mute @member1 3 days`

        """
        if not users:
            return await ctx.send_help()
        if ctx.me in users:
            return await ctx.send(_("You cannot mute me."))
        if ctx.author in users:
            return await ctx.send(_("You cannot mute yourself."))

        if not await self._check_for_mute_role(ctx):
            return
        async with ctx.typing():
            until = time_and_reason.get("until", None)
            reason = time_and_reason.get("reason", None)
            time = ""
            duration = None
            if until:
                duration = time_and_reason.get("duration")
                length = humanize_timedelta(timedelta=duration)
                time = _(" for {length} until {duration}").format(
                    length=length, duration=discord.utils.format_dt(until)
                )

            else:
                default_duration = await self.config.guild(ctx.guild).default_time()
                if default_duration:
                    duration = timedelta(seconds=default_duration)
                    until = ctx.message.created_at + duration
                    length = humanize_timedelta(seconds=default_duration)
                    time = _(" for {length} until {duration}").format(
                        length=length, duration=discord.utils.format_dt(until)
                    )

            author = ctx.message.author
            guild = ctx.guild
            audit_reason = get_audit_reason(author, reason, shorten=True)
            success_list = []
            issue_list = []
            for user in users:
                response = await self.mute_user(guild, author, user, until, audit_reason)
                if response.success:
                    success_list.append(user)
                    await modlog.create_case(
                        self.bot,
                        guild,
                        ctx.message.created_at,
                        "smute",
                        user,
                        author,
                        reason,
                        until=until,
                        channel=None,
                    )
                    await self._send_dm_notification(
                        user, author, guild, _("Server mute"), reason, duration
                    )
                else:
                    issue_list.append(response)
        if success_list:
            if ctx.guild.id not in self._server_mutes:
                self._server_mutes[ctx.guild.id] = {}
            msg = _("{users} has been muted in this server{time}.")
            if len(success_list) > 1:
                msg = _("{users} have been muted in this server{time}.")
            await ctx.send(
                msg.format(users=humanize_list([f"`{u}`" for u in success_list]), time=time)
            )
        if issue_list:
            await self.handle_issues(ctx, issue_list)

    def parse_issues(self, issues: List[Union[MuteResponse, ChannelMuteResponse]]) -> str:
        users = set(issue.user for issue in issues)
        error_msg = ""

        for user in users:
            error_msg += _("{member} could not be (un)muted for the following reasons:\n").format(
                member=f"`{user}`"
            )
            # I would like to replace this with a user mention but send_interactive
            # does not support supressing mentions at this time. So in order to keep
            # this formatting consistent the username is escaped in a code block.
            for issue in issues:
                if issue.user.id != user.id:
                    continue
                if issue.reason:
                    error_msg += f"- {issue.reason}\n"

        return error_msg

    async def handle_issues(
        self, ctx: commands.Context, issue_list: List[Union[MuteResponse, ChannelMuteResponse]]
    ) -> None:
        """
        This is to handle the various issues that can return for each user/channel
        """
        message = _(
            "Some users could not be properly muted or unmuted. Would you like to see who, where, and why?"
        )

        can_react = can_user_react_in(ctx.me, ctx.channel)
        if not can_react:
            message += " (y/n)"
        query: discord.Message = await ctx.send(message)
        if can_react:
            # noinspection PyAsyncCall
            start_adding_reactions(query, ReactionPredicate.YES_OR_NO_EMOJIS)
            pred = ReactionPredicate.yes_or_no(query, ctx.author)
            event = "reaction_add"
        else:
            pred = MessagePredicate.yes_or_no(ctx)
            event = "message"
        try:
            await ctx.bot.wait_for(event, check=pred, timeout=30)
        except asyncio.TimeoutError:
            with contextlib.suppress(discord.NotFound):
                await query.delete()
            return

        if not pred.result:
            if can_react:
                with contextlib.suppress(discord.NotFound):
                    await query.delete()
            else:
                await ctx.send(_("OK then."))
            return
        else:
            if can_react:
                with contextlib.suppress(discord.Forbidden):
                    await query.clear_reactions()
            issue = self.parse_issues(issue_list)
            resp = pagify(issue)
            await ctx.send_interactive(resp)

    @commands.command(
        name="mutechannel", aliases=["channelmute"], usage="<users...> [time_and_reason]"
    )
    @commands.mod_or_permissions(manage_roles=True)
    @commands.bot_has_guild_permissions(manage_permissions=True)
    async def channel_mute(
        self,
        ctx: commands.Context,
        users: commands.Greedy[discord.Member],
        *,
        time_and_reason: MuteTime = {},
    ):
        """Mute a user in the current text channel (or in the parent of the current thread).

        `<users...>` is a space separated list of usernames, ID's, or mentions.
        `[time_and_reason]` is the time to mute for and reason. Time is
        any valid time length such as `30 minutes` or `2 days`. If nothing
        is provided the mute will use the set default time or indefinite if not set.

        Examples:
        `[p]mutechannel @member1 @member2 spam 5 hours`
        `[p]mutechannel @member1 3 days`
        """
        if not users:
            return await ctx.send_help()
        if ctx.me in users:
            return await ctx.send(_("You cannot mute me."))
        if ctx.author in users:
            return await ctx.send(_("You cannot mute yourself."))
        async with ctx.typing():
            until = time_and_reason.get("until", None)
            reason = time_and_reason.get("reason", None)
            time = ""
            duration = None
            if until:
                duration = time_and_reason.get("duration")
                length = humanize_timedelta(timedelta=duration)
                time = _(" for {length} until {duration}").format(
                    length=length, duration=discord.utils.format_dt(until)
                )

            else:
                default_duration = await self.config.guild(ctx.guild).default_time()
                if default_duration:
                    duration = timedelta(seconds=default_duration)
                    until = ctx.message.created_at + duration
                    length = humanize_timedelta(seconds=default_duration)
                    time = _(" for {length} until {duration}").format(
                        length=length, duration=discord.utils.format_dt(until)
                    )
            author = ctx.message.author
            channel = ctx.message.channel
            if isinstance(channel, discord.Thread):
                channel = channel.parent
            guild = ctx.guild
            audit_reason = get_audit_reason(author, reason, shorten=True)
            issue_list = []
            success_list = []
            for user in users:
                response = await self.channel_mute_user(
                    guild, channel, author, user, until, audit_reason
                )
                if response.success:
                    success_list.append(user)
                    if response.reason:
                        # This is incase we couldn't move the user from voice channels
                        issue_list.append(response)
                    await modlog.create_case(
                        self.bot,
                        guild,
                        ctx.message.created_at,
                        "cmute",
                        user,
                        author,
                        reason,
                        until=until,
                        channel=channel,
                    )
                    await self._send_dm_notification(
                        user, author, guild, _("Channel mute"), reason, duration
                    )
                    async with self.config.member(user).perms_cache() as cache:
                        cache[channel.id] = response.old_overs
                else:
                    issue_list.append(response)

        if success_list:
            msg = _("{users} has been muted in this channel{time}.")
            if len(success_list) > 1:
                msg = _("{users} have been muted in this channel{time}.")
            await ctx.send(
                msg.format(users=humanize_list([f"`{u}`" for u in success_list]), time=time)
            )
        if issue_list:
            msg = _("The following users could not be muted:\n")
            for issue in issue_list:
                msg += f"- `{issue.user}`: {issue.reason}\n"
            await ctx.send_interactive(pagify(msg))

    @commands.command(usage="<users...> [reason]")
    @commands.guild_only()
    @commands.mod_or_permissions(manage_roles=True)
    async def unmute(
        self,
        ctx: commands.Context,
        users: commands.Greedy[discord.Member],
        *,
        reason: Optional[str] = None,
    ):
        """Unmute users.

        `<users...>` is a space separated list of usernames, ID's, or mentions.
        `[reason]` is the reason for the unmute.
        """
        if not users:
            return await ctx.send_help()
        if ctx.me in users:
            return await ctx.send(_("You cannot unmute me."))
        if ctx.author in users:
            return await ctx.send(_("You cannot unmute yourself."))
        if not await self._check_for_mute_role(ctx):
            return
        async with ctx.typing():
            guild = ctx.guild
            author = ctx.author
            audit_reason = get_audit_reason(author, reason, shorten=True)
            issue_list = []
            success_list = []
            if guild.id in self._channel_mute_events:
                self._channel_mute_events[guild.id].clear()
            else:
                self._channel_mute_events[guild.id] = asyncio.Event()
            for user in users:
                response = await self.unmute_user(guild, author, user, audit_reason)

                if response.success:
                    if response.reason:
                        # This is incase we couldn't move the user from voice channels
                        issue_list.append(response)
                    success_list.append(user)
                    await modlog.create_case(
                        self.bot,
                        guild,
                        ctx.message.created_at,
                        "sunmute",
                        user,
                        author,
                        reason,
                        until=None,
                    )
                    await self._send_dm_notification(
                        user, author, guild, _("Server unmute"), reason
                    )
                else:
                    issue_list.append(response)
        self._channel_mute_events[guild.id].set()
        if success_list:
            if ctx.guild.id in self._server_mutes and self._server_mutes[ctx.guild.id]:
                await self.config.guild(ctx.guild).muted_users.set(
                    self._server_mutes[ctx.guild.id]
                )
            else:
                await self.config.guild(ctx.guild).muted_users.clear()
            await ctx.send(
                _("{users} unmuted in this server.").format(
                    users=humanize_list([f"`{u}`" for u in success_list])
                )
            )
        if issue_list:
            await self.handle_issues(ctx, issue_list)

    @commands.command(usage="<users...> [reason]", hidden=True)
    @commands.guild_only()
    @commands.mod_or_permissions(manage_roles=True)
    async def forceunmute(
        self,
        ctx: commands.Context,
        users: commands.Greedy[discord.Member],
        *,
        reason: Optional[str] = None,
    ):
        """Force Unmute users who have had channel overwrite mutes in every channel.

        `<users...>` is a space separated list of usernames, ID's, or mentions.
        `[reason]` is the reason for the unmute.
        """
        if not users:
            return await ctx.send_help()
        if ctx.me in users:
            return await ctx.send(_("You cannot unmute me."))
        if ctx.author in users:
            return await ctx.send(_("You cannot unmute yourself."))
        async with ctx.typing():
            guild = ctx.guild
            author = ctx.author
            audit_reason = get_audit_reason(author, reason, shorten=True)
            issue_list = []
            success_list = []
            if guild.id in self._channel_mute_events:
                self._channel_mute_events[guild.id].clear()
            else:
                self._channel_mute_events[guild.id] = asyncio.Event()
            for user in users:
                tasks = []
                for channel in guild.channels:
                    tasks.append(
                        self.channel_unmute_user(guild, channel, author, user, audit_reason)
                    )
                results = await bounded_gather(*tasks)
                for result in results:
                    if not result.success:
                        issue_list.append(result)
                if any(t.success for t in results):
                    success_list.append(user)
                    await modlog.create_case(
                        self.bot,
                        guild,
                        ctx.message.created_at,
                        "sunmute",
                        user,
                        author,
                        reason,
                        until=None,
                    )
                    await self._send_dm_notification(
                        user, author, guild, _("Server unmute"), reason
                    )
                await self.config.member(user).clear()

        self._channel_mute_events[guild.id].set()
        if success_list:
            if ctx.guild.id in self._server_mutes and self._server_mutes[ctx.guild.id]:
                await self.config.guild(ctx.guild).muted_users.set(
                    self._server_mutes[ctx.guild.id]
                )
            else:
                await self.config.guild(ctx.guild).muted_users.clear()
            await ctx.send(
                _("{users} unmuted in this server.").format(
                    users=humanize_list([f"`{u}`" for u in success_list])
                )
            )
        if issue_list:
            await self.handle_issues(ctx, issue_list)

    @commands.mod_or_permissions(manage_roles=True)
    @commands.command(name="unmutechannel", aliases=["channelunmute"], usage="<users...> [reason]")
    @commands.bot_has_guild_permissions(manage_permissions=True)
    async def unmute_channel(
        self,
        ctx: commands.Context,
        users: commands.Greedy[discord.Member],
        *,
        reason: Optional[str] = None,
    ):
        """Unmute a user in this channel (or in the parent of this thread).

        `<users...>` is a space separated list of usernames, ID's, or mentions.
        `[reason]` is the reason for the unmute.
        """
        if not users:
            return await ctx.send_help()
        if ctx.me in users:
            return await ctx.send(_("You cannot unmute me."))
        if ctx.author in users:
            return await ctx.send(_("You cannot unmute yourself."))
        async with ctx.typing():
            channel = ctx.channel
            if isinstance(channel, discord.Thread):
                channel = channel.parent
            author = ctx.author
            guild = ctx.guild
            audit_reason = get_audit_reason(author, reason, shorten=True)
            success_list = []
            issue_list = []
            for user in users:
                response = await self.channel_unmute_user(
                    guild, channel, author, user, audit_reason
                )

                if response.success:
                    success_list.append(user)
                    await modlog.create_case(
                        self.bot,
                        guild,
                        ctx.message.created_at,
                        "cunmute",
                        user,
                        author,
                        reason,
                        until=None,
                        channel=channel,
                    )
                    await self._send_dm_notification(
                        user, author, guild, _("Channel unmute"), reason
                    )
                else:
                    issue_list.append(response)
        if success_list:
            if channel.id in self._channel_mutes and self._channel_mutes[channel.id]:
                await self.config.channel(channel).muted_users.set(self._channel_mutes[channel.id])
            else:
                await self.config.channel(channel).muted_users.clear()
            await ctx.send(
                _("{users} unmuted in this channel.").format(
                    users=humanize_list([f"`{u}`" for u in success_list])
                )
            )
        if issue_list:
            msg = _("The following users could not be unmuted:\n")
            for issue in issue_list:
                msg += f"- `{issue.user}`: {issue.reason}\n"
            await ctx.send_interactive(pagify(msg))

    async def mute_user(
        self,
        guild: discord.Guild,
        author: discord.Member,
        user: discord.Member,
        until: Optional[datetime] = None,
        reason: Optional[str] = None,
    ) -> MuteResponse:
        """
        Handles muting users
        """
        permissions = user.guild_permissions
        ret: MuteResponse = MuteResponse(success=False, reason=None, user=user)

        if permissions.administrator:
            ret.reason = _(MUTE_UNMUTE_ISSUES["is_admin"])
            return ret
        if not await self.is_allowed_by_hierarchy(guild, author, user):
            ret.reason = _(MUTE_UNMUTE_ISSUES["hierarchy_problem"])
            return ret
        mute_role = await self.config.guild(guild).mute_role()

        if mute_role:
            role = guild.get_role(mute_role)
            if not role:
                ret.reason = _(MUTE_UNMUTE_ISSUES["role_missing"])
                return ret
            if author != guild.owner and role >= author.top_role:
                ret.reason = _(MUTE_UNMUTE_ISSUES["assigned_role_hierarchy_problem"])
                return ret
            if not guild.me.guild_permissions.manage_roles or role >= guild.me.top_role:
                ret.reason = _(MUTE_UNMUTE_ISSUES["permissions_issue_role"])
                return ret
            # This is here to prevent the modlog case from happening on role updates
            # we need to update the cache early so it's there before we receive the member_update event
            if guild.id not in self._server_mutes:
                self._server_mutes[guild.id] = {}

            self._server_mutes[guild.id][user.id] = {
                "author": author.id,
                "member": user.id,
                "until": until.timestamp() if until else None,
            }
            try:
                await user.add_roles(role, reason=reason)
                await self.config.guild(guild).muted_users.set(self._server_mutes[guild.id])
            except discord.errors.Forbidden:
                if guild.id in self._server_mutes and user.id in self._server_mutes[guild.id]:
                    del self._server_mutes[guild.id][user.id]
                ret.reason = _(MUTE_UNMUTE_ISSUES["permissions_issue_role"])
                return ret
            if user.voice:
                try:
                    await user.move_to(user.voice.channel)
                except discord.HTTPException:
                    # catch all discord errors because the result will be the same
                    # we successfully muted by this point but can't move the user
                    ret.reason = _(MUTE_UNMUTE_ISSUES["voice_mute_permission"])
            ret.success = True
            return ret
        else:
            if until and (until - datetime.now(tz=timezone.utc)) > timedelta(days=28):
                ret.reason = _(MUTE_UNMUTE_ISSUES["mute_is_too_long"])
                return ret
            if not until:
                ret.reason = _(MUTE_UNMUTE_ISSUES["timeouts_require_time"])
                return ret
            if guild.me.guild_permissions.moderate_members:
                try:
                    await user.edit(timed_out_until=until, reason=reason)
                    ret.success = True
                except Exception:
                    ret.reason = _(MUTE_UNMUTE_ISSUES["permissions_issue_guild"])
            else:
                ret.reason = _("I lack the moderate members permission.")
            return ret

    async def unmute_user(
        self,
        guild: discord.Guild,
        author: Optional[discord.Member],
        user: discord.Member,
        reason: Optional[str] = None,
    ) -> MuteResponse:
        """
        Handles unmuting users
        """
        ret: MuteResponse = MuteResponse(success=False, reason=None, user=user)

        mute_role_id = await self.config.guild(guild).mute_role()
        if author is not None and not await self.is_allowed_by_hierarchy(guild, author, user):
            ret.reason = _(MUTE_UNMUTE_ISSUES["hierarchy_problem"])
            return ret

        reasons = []
        mute_role = guild.get_role(mute_role_id)

        if mute_role and mute_role in user.roles:
            if guild.id in self._server_mutes:
                if user.id in self._server_mutes[guild.id]:
                    del self._server_mutes[guild.id][user.id]
            if not guild.me.guild_permissions.manage_roles or mute_role >= guild.me.top_role:
                reasons.append(_(MUTE_UNMUTE_ISSUES["permissions_issue_role"]))
            else:
                try:
                    await user.remove_roles(mute_role, reason=reason)
                    ret.success = True
                except discord.errors.Forbidden:
                    reasons.append(_(MUTE_UNMUTE_ISSUES["permissions_issue_role"]))

        if user.is_timed_out():
            if guild.me.guild_permissions.moderate_members:
                try:
                    await user.edit(timed_out_until=None, reason=reason)
                    ret.success = True
                except Exception:
                    reasons.append(_(MUTE_UNMUTE_ISSUES["permissions_issue_guild"]))
            else:
                reasons.append(_("I lack the timeout members permission."))

        if not reasons and not ret.success:
            ret.reason = _(MUTE_UNMUTE_ISSUES["already_unmuted"]).format(location=_("this server"))
        elif reasons:
            ret.reason = "\n".join(reasons)

        return ret

    async def channel_mute_user(
        self,
        guild: discord.Guild,
        channel: discord.abc.GuildChannel,
        author: discord.Member,
        user: discord.Member,
        until: Optional[datetime] = None,
        reason: Optional[str] = None,
        *,
        voice_mute: bool = False,
    ) -> ChannelMuteResponse:
        """Mutes the specified user in the specified channel"""
        overwrites = channel.overwrites_for(user)
        permissions = channel.permissions_for(user)

        ret = ChannelMuteResponse(
            success=False,
            channel=channel,
            reason=None,
            user=user,
            old_overs={},
            voice_mute=voice_mute,
        )

        if permissions.administrator:
            ret.reason = _(MUTE_UNMUTE_ISSUES["is_admin"])
            return ret

        move_channel = False
        if user.voice and user.voice.channel == channel:
            if channel.permissions_for(guild.me).move_members:
                move_channel = True
            else:
                ret.reason = _(MUTE_UNMUTE_ISSUES["voice_mute_permission"])

        if not await self.is_allowed_by_hierarchy(guild, author, user):
            ret.reason = _(MUTE_UNMUTE_ISSUES["hierarchy_problem"])
            return ret

        if channel.id not in self._channel_mutes:
            self._channel_mutes[channel.id] = {}
        current_mute = self._channel_mutes[channel.id].get(user.id)

        # Determine if this is voice mute -> channel mute upgrade
        is_mute_upgrade = (
            current_mute is not None and not voice_mute and current_mute.get("voice_mute", False)
        )
        # We want to continue if this is a new mute or a mute upgrade,
        # otherwise we should return with failure.
        if current_mute is not None and not is_mute_upgrade:
            ret.reason = _(MUTE_UNMUTE_ISSUES["already_muted"]).format(location=channel.mention)
            return ret
        new_overs: Dict[str, Optional[bool]] = {"speak": False}
        if not voice_mute:
            new_overs.update(
                send_messages=False,
                send_messages_in_threads=False,
                create_public_threads=False,
                create_private_threads=False,
                use_application_commands=False,
                add_reactions=False,
            )
        old_overs = {k: getattr(overwrites, k) for k in new_overs}
        if is_mute_upgrade:
            perms_cache = await self.config.member(user).perms_cache()
            if "speak" in perms_cache:
                old_overs["speak"] = perms_cache["speak"]
        ret.old_overs = old_overs
        overwrites.update(**new_overs)
        if not channel.permissions_for(guild.me).manage_permissions:
            ret.reason = _(MUTE_UNMUTE_ISSUES["permissions_issue_channel"]).format(
                location=channel.mention
            )
            return ret

        self._channel_mutes[channel.id][user.id] = {
            "author": author.id,
            "guild": guild.id,
            "member": user.id,
            "until": until.timestamp() if until else None,
            "voice_mute": voice_mute,
        }
        try:
            await channel.set_permissions(user, overwrite=overwrites, reason=reason)
            async with self.config.channel(channel).muted_users() as muted_users:
                muted_users[str(user.id)] = self._channel_mutes[channel.id][user.id]
        except discord.NotFound as e:
            if channel.id in self._channel_mutes and user.id in self._channel_mutes[channel.id]:
                del self._channel_mutes[channel.id][user.id]
            if e.code == 10003:
                if (
                    channel.id in self._channel_mutes
                    and user.id in self._channel_mutes[channel.id]
                ):
                    del self._channel_mutes[channel.id][user.id]
                ret.reason = _(MUTE_UNMUTE_ISSUES["unknown_channel"])
                return ret

            elif e.code == 10009:
                if (
                    channel.id in self._channel_mutes
                    and user.id in self._channel_mutes[channel.id]
                ):
                    del self._channel_mutes[channel.id][user.id]
                ret.reason = _(MUTE_UNMUTE_ISSUES["left_guild"])
                return ret

        except discord.Forbidden:
            ret.reason = _(MUTE_UNMUTE_ISSUES["permissions_issue_channel"]).format(
                location=channel.mention
            )
            return ret

        if move_channel:
            try:
                await user.move_to(channel)
            except discord.HTTPException:
                # catch all discord errors because the result will be the same
                # we successfully muted by this point but can't move the user
                ret.reason = _(MUTE_UNMUTE_ISSUES["voice_mute_permission"])
                ret.success = True
                return ret
        ret.success = True
        return ret

    async def channel_unmute_user(
        self,
        guild: discord.Guild,
        channel: discord.abc.GuildChannel,
        author: Optional[discord.Member],
        user: discord.Member,
        reason: Optional[str] = None,
        *,
        voice_mute: bool = False,
    ) -> ChannelMuteResponse:
        """Unmutes the specified user in a specified channel"""
        overwrites = channel.overwrites_for(user)
        perms_cache = await self.config.member(user).perms_cache()

        ret = ChannelMuteResponse(
            success=False,
            reason=None,
            user=user,
            channel=channel,
            old_overs={},
            voice_mute=voice_mute,
        )

        move_channel = False
        if channel.id in perms_cache:
            old_values = perms_cache[channel.id]
        else:
            old_values = {
                "send_messages": None,
                "send_messages_in_threads": None,
                "create_public_threads": None,
                "create_private_threads": None,
                "use_application_commands": None,
                "add_reactions": None,
                "speak": None,
            }

        if user.voice and user.voice.channel == channel:
            if channel.permissions_for(guild.me).move_members:
                move_channel = True

        if author is not None and not await self.is_allowed_by_hierarchy(guild, author, user):
            ret.reason = _(MUTE_UNMUTE_ISSUES["hierarchy_problem"])
            return ret

        overwrites.update(**old_values)
        if channel.id in self._channel_mutes and user.id in self._channel_mutes[channel.id]:
            current_mute = self._channel_mutes[channel.id].pop(user.id)
        else:
            ret.reason = _(MUTE_UNMUTE_ISSUES["already_unmuted"]).format(location=channel.mention)
            return ret

        if not current_mute["voice_mute"] and voice_mute:
            ret.reason = _(MUTE_UNMUTE_ISSUES["is_not_voice_mute"]).format(
                command=inline("unmutechannel")
            )
            return ret

        if not channel.permissions_for(guild.me).manage_permissions:
            ret.reason = _(MUTE_UNMUTE_ISSUES["permissions_issue_channel"]).format(
                location=channel.mention
            )
            return ret

        try:
            if overwrites.is_empty():
                await channel.set_permissions(
                    user, overwrite=cast(discord.PermissionOverwrite, None), reason=reason
                )
            else:
                await channel.set_permissions(user, overwrite=overwrites, reason=reason)
            async with self.config.channel(channel).muted_users() as muted_users:
                if str(user.id) in muted_users:
                    del muted_users[str(user.id)]
        except discord.NotFound as e:
            if e.code == 10003:
                ret.reason = _(MUTE_UNMUTE_ISSUES["unknown_channel"])
                return ret

            elif e.code == 10009:
                ret.reason = _(MUTE_UNMUTE_ISSUES["left_guild"])
                return ret

        if move_channel:
            try:
                await user.move_to(channel)
            except discord.HTTPException:
                # catch all discord errors because the result will be the same
                # we successfully muted by this point but can't move the user
                ret.success = True
                ret.reason = _(MUTE_UNMUTE_ISSUES["voice_mute_permission"])
                return ret
        ret.success = True
        return ret



from __future__ import annotations

import getpass
import os
import platform
import sys
from typing import Optional

import discord
import pip
import psutil

from redbot import __version__
from redbot.core import data_manager
from redbot.core.bot import Red
from redbot.core.utils.chat_formatting import box


def noop_box(text: str, **kwargs) -> str:
    return text


def _datasize(num: int):
    for unit in ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB"]:
        if abs(num) < 1024.0:
            return "{0:.1f}{1}".format(num, unit)
        num /= 1024.0
    return "{0:.1f}{1}".format(num, "YB")


class DebugInfoSection:
    def __init__(self, section_name: str, *section_parts: str) -> None:
        self.section_name = section_name
        self.section_parts = section_parts

    def get_command_text(self) -> str:
        parts = [box(f"## {self.section_name}:", lang="md")]
        for part in self.section_parts:
            parts.append(box(part))
        return "".join(parts)

    def get_cli_text(self) -> str:
        parts = [f"\x1b[32m## {self.section_name}:\x1b[0m"]
        for part in self.section_parts:
            parts.append(part)
        return "\n".join(parts)


class DebugInfo:
    def __init__(self, bot: Optional[Red] = None) -> None:
        self.bot = bot

    @property
    def is_logged_in(self) -> bool:
        return self.bot is not None and self.bot.application_id is not None

    @property
    def is_connected(self) -> bool:
        return self.bot is not None and self.bot.is_ready()

    async def get_cli_text(self) -> str:
        parts = ["\x1b[31m# Debug Info for Red:\x1b[0m"]
        for section in (
            self._get_system_metadata_section(),
            self._get_os_variables_section(),
            await self._get_red_vars_section(),
        ):
            parts.append("")
            parts.append(section.get_cli_text())

        return "\n".join(parts)

    async def get_command_text(self) -> str:
        parts = [box("# Debug Info for Red:", lang="md")]
        for section in (
            self._get_system_metadata_section(),
            self._get_os_variables_section(),
            await self._get_red_vars_section(),
        ):
            parts.append("\n")
            parts.append(section.get_command_text())

        return "".join(parts)

    def _get_system_metadata_section(self) -> DebugInfoSection:
        memory_ram = psutil.virtual_memory()
        ram_string = "{used}/{total} ({percent}%)".format(
            used=_datasize(memory_ram.used),
            total=_datasize(memory_ram.total),
            percent=memory_ram.percent,
        )
        return DebugInfoSection(
            "System Metadata",
            f"CPU Cores: {psutil.cpu_count()} ({platform.machine()})\nRAM: {ram_string}",
        )

    def _get_os_variables_section(self) -> DebugInfoSection:
        IS_WINDOWS = os.name == "nt"
        IS_MAC = sys.platform == "darwin"
        IS_LINUX = sys.platform == "linux"

        python_version = ".".join(map(str, sys.version_info[:3]))
        pyver = f"{python_version} ({platform.architecture()[0]})"
        pipver = pip.__version__
        redver = __version__
        dpy_version = discord.__version__
        if IS_WINDOWS:
            os_info = platform.uname()
            osver = f"{os_info.system} {os_info.release} (version {os_info.version})"
        elif IS_MAC:
            os_info = platform.mac_ver()
            osver = f"Mac OSX {os_info[0]} {os_info[2]}"
        elif IS_LINUX:
            import distro

            osver = f"{distro.name()} {distro.version()}".strip()
        else:
            osver = "Could not parse OS, report this on Github."
        user_who_ran = getpass.getuser()

        resp_os = f"OS version: {osver}\nUser: {user_who_ran}\n"  # Ran where off to?!
        resp_py_metadata = (
            f"Python executable: {sys.executable}\n"
            f"Python version: {pyver}\n"
            f"Pip version: {pipver}\n"
        )
        resp_red_metadata = f"Red version: {redver}\nDiscord.py version: {dpy_version}"
        return DebugInfoSection(
            "OS variables",
            resp_os,
            resp_py_metadata,
            resp_red_metadata,
        )

    async def _get_red_vars_section(self) -> DebugInfoSection:
        instance_name = data_manager.instance_name()
        if instance_name is None:
            return DebugInfoSection(
                "Red variables",
                f"Metadata file: {data_manager.config_file}",
            )

        parts = [f"Instance name: {instance_name}"]

        if self.bot is not None:
            # sys.original_argv is available since 3.10 and shows the actual command line arguments
            # rather than a Python-transformed version (i.e. with '-c' or path to `__main__.py`
            # as first element). We could just not show the first argument for consistency
            # but it can be useful.
            cli_args = getattr(sys, "orig_argv", sys.argv).copy()
            # best effort attempt to expunge a token argument
            for idx, arg in enumerate(cli_args):
                if not arg.startswith("--to"):
                    continue
                arg_name, sep, arg_value = arg.partition("=")
                if arg_name not in ("--to", "--tok", "--toke", "--token"):
                    continue
                if sep:
                    cli_args[idx] = f"{arg_name}{sep}[EXPUNGED]"
                elif len(cli_args) > idx + 1:
                    cli_args[idx + 1] = f"[EXPUNGED]"
            parts.append(f"Command line arguments: {cli_args!r}")

            # This formatting is a bit ugly but this is a debug information command
            # and calling repr() on prefix strings ensures that the list isn't ambiguous.
            prefixes = ", ".join(map(repr, await self.bot._config.prefix()))
            parts.append(f"Global prefix(es): {prefixes}")

        if self.is_logged_in:
            owners = []
            for uid in self.bot.owner_ids:
                try:
                    u = await self.bot.get_or_fetch_user(uid)
                    owners.append(f"{u.id} ({u})")
                except discord.HTTPException:
                    owners.append(f"{uid} (Unresolvable)")
            owners_string = ", ".join(owners) or "None"
            parts.append(f"Owner(s): {', '.join(owners) or 'None'}")

        if self.is_connected:
            disabled_intents = (
                ", ".join(
                    intent_name.replace("_", " ").title()
                    for intent_name, enabled in self.bot.intents
                    if not enabled
                )
                or "None"
            )
            parts.append(f"Disabled intents: {disabled_intents}")

        parts.append(f"Storage type: {data_manager.storage_type()}")
        parts.append(f"Data path: {data_manager.basic_config['DATA_PATH']}")
        parts.append(f"Metadata file: {data_manager.config_file}")

        return DebugInfoSection(
            "Red variables",
            "\n".join(parts),
        )



from typing import Final, Tuple

from .ll_version import LavalinkVersion

__all__ = (
    "JAR_VERSION",
    "YT_PLUGIN_VERSION",
    "SUPPORTED_JAVA_VERSIONS",
    "LATEST_SUPPORTED_JAVA_VERSION",
    "OLDER_SUPPORTED_JAVA_VERSIONS",
)


JAR_VERSION: Final[LavalinkVersion] = LavalinkVersion(3, 7, 12, red=1)
YT_PLUGIN_VERSION: Final[str] = "1.11.4"
# keep this sorted from oldest to latest
SUPPORTED_JAVA_VERSIONS: Final[Tuple[int, ...]] = (11, 17)
LATEST_SUPPORTED_JAVA_VERSION: Final = SUPPORTED_JAVA_VERSIONS[-1]
OLDER_SUPPORTED_JAVA_VERSIONS: Final[Tuple[int, ...]] = SUPPORTED_JAVA_VERSIONS[:-1]



########## SENSITIVE SECTION WARNING ###########
################################################
# Any edits of any of the exported names       #
# may result in a breaking change.             #
# Ensure no names are removed without warning. #
################################################

### DEP-WARN: Check this *every* discord.py update
from discord.app_commands.checks import (
    bot_has_permissions,
    cooldown,
    dynamic_cooldown,
    has_any_role,
    has_role,
    has_permissions,
)

__all__ = (
    "bot_has_permissions",
    "cooldown",
    "dynamic_cooldown",
    "has_any_role",
    "has_role",
    "has_permissions",
)



import datetime
import heapq
import math
import random
from pathlib import Path

import discord
import lavalink
from red_commons.logging import getLogger

from redbot.core import commands
from redbot.core.i18n import Translator
from redbot.core.utils import AsyncIter
from redbot.core.utils.chat_formatting import humanize_number, pagify
from redbot.core.utils.menus import menu

from ..abc import MixinMeta
from ..cog_utils import CompositeMetaClass

log = getLogger("red.cogs.Audio.cog.Commands.miscellaneous")
_ = Translator("Audio", Path(__file__))


class MiscellaneousCommands(MixinMeta, metaclass=CompositeMetaClass):
    @commands.command(name="sing")
    @commands.guild_only()
    @commands.bot_has_permissions(embed_links=True)
    async def command_sing(self, ctx: commands.Context):
        """Make Red sing one of her songs."""
        ids = (
            "zGTkAVsrfg8",
            "cGMWL8cOeAU",
            "vFrjMq4aL-g",
            "WROI5WYBU_A",
            "41tIUr_ex3g",
            "f9O2Rjn1azc",
        )
        url = f"https://www.youtube.com/watch?v={random.choice(ids)}"
        await ctx.invoke(self.command_play, query=url)

    @commands.command(name="audiostats")
    @commands.guild_only()
    @commands.is_owner()
    @commands.bot_has_permissions(embed_links=True)
    @commands.bot_can_react()
    async def command_audiostats(self, ctx: commands.Context):
        """Audio stats."""
        server_num = len(lavalink.active_players())
        total_num = len(lavalink.all_connected_players())

        msg = ""
        async for p in AsyncIter(lavalink.all_connected_players()):
            connect_dur = (
                self.get_time_string(
                    int(
                        (
                            datetime.datetime.now(datetime.timezone.utc) - p.connected_at
                        ).total_seconds()
                    )
                )
                or "0s"
            )
            try:
                if not p.current:
                    raise AttributeError
                current_title = await self.get_track_description(
                    p.current, self.local_folder_current_path
                )
                msg += f"{p.guild.name} [`{connect_dur}`]: {current_title}\n"
            except AttributeError:
                msg += "{} [`{}`]: **{}**\n".format(
                    p.guild.name, connect_dur, _("Nothing playing.")
                )

        if total_num == 0:
            return await self.send_embed_msg(ctx, title=_("Not connected anywhere."))
        servers_embed = []
        pages = 1
        for page in pagify(msg, delims=["\n"], page_length=1500):
            em = discord.Embed(
                colour=await ctx.embed_colour(),
                title=_("Playing in {num}/{total} servers:").format(
                    num=humanize_number(server_num), total=humanize_number(total_num)
                ),
                description=page,
            )
            em.set_footer(
                text=_("Page {}/{}").format(
                    humanize_number(pages), humanize_number((math.ceil(len(msg) / 1500)))
                )
            )
            pages += 1
            servers_embed.append(em)

        await menu(ctx, servers_embed)

    @commands.command(name="percent")
    @commands.guild_only()
    @commands.bot_has_permissions(embed_links=True)
    async def command_percent(self, ctx: commands.Context):
        """Queue percentage."""
        if not self._player_check(ctx):
            return await self.send_embed_msg(ctx, title=_("Nothing playing."))
        player = lavalink.get_player(ctx.guild.id)
        queue_tracks = player.queue
        requesters = {"total": 0, "users": {}}

        async def _usercount(req_user_handle):
            if req_user_handle in requesters["users"]:
                requesters["users"][req_user_handle]["songcount"] += 1
                requesters["total"] += 1
            else:
                requesters["users"][req_user_handle] = {}
                requesters["users"][req_user_handle]["songcount"] = 1
                requesters["total"] += 1

        async for track in AsyncIter(queue_tracks):
            req_user_handle = str(track.requester)
            await _usercount(req_user_handle)

        try:
            req_user_handle = str(player.current.requester)
            await _usercount(req_user_handle)
        except AttributeError:
            return await self.send_embed_msg(ctx, title=_("There's nothing in the queue."))

        async for req_user_handle in AsyncIter(requesters["users"]):
            percentage = float(requesters["users"][req_user_handle]["songcount"]) / float(
                requesters["total"]
            )
            requesters["users"][req_user_handle]["percent"] = round(percentage * 100, 1)

        top_queue_users = heapq.nlargest(
            20,
            [
                (x, requesters["users"][x][y])
                for x in requesters["users"]
                for y in requesters["users"][x]
                if y == "percent"
            ],
            key=lambda x: x[1],
        )
        queue_user = ["{}: {:g}%".format(x[0], x[1]) for x in top_queue_users]
        queue_user_list = "\n".join(queue_user)
        await self.send_embed_msg(
            ctx, title=_("Queued and playing tracks:"), description=queue_user_list
        )



"""
The original implementation of this cog was heavily based on
RoboDanny's REPL cog which can be found here:
https://github.com/Rapptz/RoboDanny/blob/f13e1c9a6a7205e50de6f91fa5326fc7113332d3/cogs/repl.py

Copyright (c) 2017-present Cog Creators
Copyright (c) 2016-2017 Rapptz

The original copy was distributed under MIT License and this derivative work
is distributed under GNU GPL Version 3.
"""

from __future__ import annotations

import ast
import asyncio
import aiohttp
import inspect
import io
import textwrap
import traceback
import types
import re
import sys
from copy import copy
from typing import Any, Awaitable, Dict, Iterator, List, Literal, Tuple, Type, TypeVar, Union
from types import CodeType, TracebackType

import discord

from . import commands
from .commands import NoParseOptional as Optional
from .i18n import Translator, cog_i18n
from .utils import chat_formatting
from .utils.chat_formatting import pagify
from .utils.predicates import MessagePredicate

_ = Translator("Dev", __file__)

# we want to match either:
# - "```lang\n"
# - or "```" and potentially also strip a single "\n" if it follows it immediately
START_CODE_BLOCK_RE = re.compile(r"^((```[\w.+\-]+\n+(?!```))|(```\n*))")

T = TypeVar("T")


def get_pages(msg: str) -> Iterator[str]:
    """Pagify the given message for output to the user."""
    return pagify(msg, delims=["\n", " "], priority=True, shorten_by=10)


def sanitize_output(ctx: commands.Context, to_sanitize: str) -> str:
    """Hides the bot's token from a string."""
    token = ctx.bot.http.token
    if token:
        return re.sub(re.escape(token), "[EXPUNGED]", to_sanitize, re.I)
    return to_sanitize


def async_compile(source: str, filename: str, mode: Literal["eval", "exec"]) -> CodeType:
    return compile(
        source, filename, mode, flags=ast.PyCF_ALLOW_TOP_LEVEL_AWAIT, optimize=0, dont_inherit=True
    )


async def maybe_await(coro: Union[T, Awaitable[T], Awaitable[Awaitable[T]]]) -> T:
    for i in range(2):
        if inspect.isawaitable(coro):
            coro = await coro
        else:
            break
    return coro  # type: ignore


def cleanup_code(content: str) -> str:
    """Automatically removes code blocks from the code."""
    # remove ```py\n```
    if content.startswith("```") and content.endswith("```"):
        return START_CODE_BLOCK_RE.sub("", content)[:-3].rstrip("\n")

    # remove `foo`
    return content.strip("` \n")


class SourceCache:
    MAX_SIZE = 1000

    def __init__(self) -> None:
        # estimated to take less than 100 kB
        self._data: Dict[str, Tuple[str, int]] = {}
        # this just keeps going up until the bot is restarted, shouldn't really be an issue
        self._next_index = 0

    def take_next_index(self) -> int:
        next_index = self._next_index
        self._next_index += 1
        return next_index

    def __getitem__(self, key: str) -> Tuple[List[str], int]:
        value = self._data.pop(key)  # pop to put it at the end as most recent
        self._data[key] = value
        # To mimic linecache module's behavior,
        # all lines (including the last one) should end with \n.
        source_lines = [f"{line}\n" for line in value[0].splitlines()]
        # Note: while it might seem like a waste of time to always calculate the list of source lines,
        # this is a necessary memory optimization. If all of the data in `self._data` were list,
        # it could theoretically take up to 1000x as much memory.
        return source_lines, value[1]

    def __setitem__(self, key: str, value: Tuple[str, int]) -> None:
        self._data.pop(key, None)
        self._data[key] = value
        if len(self._data) > self.MAX_SIZE:
            del self._data[next(iter(self._data))]


class DevOutput:
    def __init__(
        self,
        ctx: commands.Context,
        *,
        source_cache: SourceCache,
        filename: str,
        source: str,
        env: Dict[str, Any],
    ) -> None:
        self.ctx = ctx
        self.source_cache = source_cache
        self.filename = filename
        self.source_line_offset = 0
        #: raw source - as received from the command after stripping the code block
        self.raw_source = source
        self.set_compilable_source(source)
        self.env = env
        self.always_include_result = False
        self._stream = io.StringIO()
        self.formatted_exc = ""
        self.result: Any = None
        self._old_streams = []

    @property
    def compilable_source(self) -> str:
        """Source string that we pass to async_compile()."""
        return self._compilable_source

    def set_compilable_source(self, compilable_source: str, *, line_offset: int = 0) -> None:
        self._compilable_source = compilable_source
        self.source_line_offset = line_offset
        self.source_cache[self.filename] = (compilable_source, line_offset)

    def __str__(self) -> str:
        output = []
        printed = self._stream.getvalue()
        if printed:
            output.append(printed)
        if self.formatted_exc:
            output.append(self.formatted_exc)
        elif self.always_include_result or self.result is not None:
            try:
                output.append(str(self.result))
            except Exception as exc:
                output.append(self.format_exception(exc))
        return sanitize_output(self.ctx, "".join(output))

    async def send(self, *, tick: bool = True) -> None:
        await self.ctx.send_interactive(get_pages(str(self)), box_lang="py")
        if tick and not self.formatted_exc:
            await self.ctx.tick()

    def set_exception(self, exc: Exception, *, skip_frames: int = 1) -> None:
        self.formatted_exc = self.format_exception(exc, skip_frames=skip_frames)

    def __enter__(self) -> None:
        self._old_streams.append(sys.stdout)
        sys.stdout = self._stream

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_tb: Optional[TracebackType],
        /,
    ) -> None:
        sys.stdout = self._old_streams.pop()

    @classmethod
    async def from_debug(
        cls, ctx: commands.Context, *, source: str, source_cache: SourceCache, env: Dict[str, Any]
    ) -> DevOutput:
        output = cls(
            ctx,
            source=source,
            source_cache=source_cache,
            filename=f"<debug command - snippet #{source_cache.take_next_index()}>",
            env=env,
        )
        await output.run_debug()
        return output

    @classmethod
    async def from_eval(
        cls, ctx: commands.Context, *, source: str, source_cache: SourceCache, env: Dict[str, Any]
    ) -> DevOutput:
        output = cls(
            ctx,
            source=source,
            source_cache=source_cache,
            filename=f"<eval command - snippet #{source_cache.take_next_index()}>",
            env=env,
        )
        await output.run_eval()
        return output

    @classmethod
    async def from_repl(
        cls, ctx: commands.Context, *, source: str, source_cache: SourceCache, env: Dict[str, Any]
    ) -> DevOutput:
        output = cls(
            ctx,
            source=source,
            source_cache=source_cache,
            filename=f"<repl session - snippet #{source_cache.take_next_index()}>",
            env=env,
        )
        await output.run_repl()
        return output

    async def run_debug(self) -> None:
        self.always_include_result = True
        self.set_compilable_source(self.raw_source)
        try:
            compiled = self.async_compile_with_eval()
        except SyntaxError as exc:
            self.set_exception(exc, skip_frames=3)
            return

        try:
            self.result = await maybe_await(eval(compiled, self.env))
        except Exception as exc:
            self.set_exception(exc)

    async def run_eval(self) -> None:
        self.always_include_result = False
        self.set_compilable_source(
            "async def func():\n%s" % textwrap.indent(self.raw_source, "  "), line_offset=1
        )
        try:
            compiled = self.async_compile_with_exec()
            exec(compiled, self.env)
        except SyntaxError as exc:
            self.set_exception(exc, skip_frames=3)
            return

        func = self.env["func"]
        try:
            with self:
                self.result = await func()
        except Exception as exc:
            self.set_exception(exc)

    async def run_repl(self) -> None:
        self.always_include_result = False
        self.set_compilable_source(self.raw_source)
        executor = None
        if self.raw_source.count("\n") == 0:
            # single statement, potentially 'eval'
            try:
                code = self.async_compile_with_eval()
            except SyntaxError:
                pass
            else:
                executor = eval

        if executor is None:
            try:
                code = self.async_compile_with_exec()
            except SyntaxError as exc:
                self.set_exception(exc, skip_frames=3)
                return

        try:
            with self:
                if executor is None:
                    result = types.FunctionType(code, self.env)()
                else:
                    result = executor(code, self.env)
                self.result = await maybe_await(result)
        except Exception as exc:
            self.set_exception(exc)
        else:
            if self.result is not None:
                self.env["_"] = self.result

    def async_compile_with_exec(self) -> CodeType:
        return async_compile(self.compilable_source, self.filename, "exec")

    def async_compile_with_eval(self) -> CodeType:
        return async_compile(self.compilable_source, self.filename, "eval")

    def format_exception(self, exc: Exception, *, skip_frames: int = 1) -> str:
        """
        Format an exception to send to the user.

        This function makes a few alterations to the traceback:
        - First `skip_frames` frames are skipped so that we don't show the frames
          that are part of Red's code to the user
        - `FrameSummary` objects that we get from traceback module are updated
          with the string for the corresponding line of code as otherwise
          the generated traceback string wouldn't show user's code.
        - If `line_offset` is passed, this function subtracts it from line numbers
          in `FrameSummary` objects so that those numbers properly correspond to
          the code that was provided by the user. This is needed for cases where
          we wrap user's code in an async function before exec-ing it.
        """
        exc_type = type(exc)
        tb = exc.__traceback__
        for x in range(skip_frames):
            if tb is None:
                break
            tb = tb.tb_next

        filename = self.filename
        # sometimes SyntaxError.text is None, sometimes it isn't
        if issubclass(exc_type, SyntaxError) and exc.lineno is not None:
            try:
                source_lines, line_offset = self.source_cache[exc.filename]
            except KeyError:
                pass
            else:
                if exc.text is None:
                    try:
                        # line numbers are 1-based, the list indexes are 0-based
                        exc.text = source_lines[exc.lineno - 1]
                    except IndexError:
                        # the frame might be pointing at a different source code, ignore...
                        pass
                    else:
                        exc.lineno -= line_offset
                        if sys.version_info >= (3, 10) and exc.end_lineno is not None:
                            exc.end_lineno -= line_offset
                else:
                    exc.lineno -= line_offset
                    if sys.version_info >= (3, 10) and exc.end_lineno is not None:
                        exc.end_lineno -= line_offset

        top_traceback_exc = traceback.TracebackException(exc_type, exc, tb)
        py311_or_above = sys.version_info >= (3, 11)
        queue = [  # actually a stack but 'stack' is easy to confuse with actual traceback stack
            top_traceback_exc,
        ]
        seen = {id(top_traceback_exc)}
        while queue:
            traceback_exc = queue.pop()

            # handle exception groups; this uses getattr() to support `exceptiongroup` backport lib
            exceptions: List[traceback.TracebackException] = (
                getattr(traceback_exc, "exceptions", None) or []
            )
            # handle exception chaining
            if traceback_exc.__cause__ is not None:
                exceptions.append(traceback_exc.__cause__)
            if traceback_exc.__context__ is not None:
                exceptions.append(traceback_exc.__context__)
            for te in exceptions:
                if id(te) not in seen:
                    queue.append(te)
                    seen.add(id(te))

            stack_summary = traceback_exc.stack
            for idx, frame_summary in enumerate(stack_summary):
                try:
                    source_lines, line_offset = self.source_cache[frame_summary.filename]
                except KeyError:
                    continue
                lineno = frame_summary.lineno
                if lineno is None:
                    continue

                try:
                    # line numbers are 1-based, the list indexes are 0-based
                    line = source_lines[lineno - 1]
                except IndexError:
                    # the frame might be pointing at a different source code, ignore...
                    continue
                lineno -= line_offset
                # support for enhanced error locations in tracebacks
                if py311_or_above:
                    end_lineno = frame_summary.end_lineno
                    if end_lineno is not None:
                        end_lineno -= line_offset
                    frame_summary = traceback.FrameSummary(
                        frame_summary.filename,
                        lineno,
                        frame_summary.name,
                        line=line,
                        end_lineno=end_lineno,
                        colno=frame_summary.colno,
                        end_colno=frame_summary.end_colno,
                    )
                else:
                    frame_summary = traceback.FrameSummary(
                        frame_summary.filename, lineno, frame_summary.name, line=line
                    )
                stack_summary[idx] = frame_summary

        return "".join(top_traceback_exc.format())


@cog_i18n(_)
class Dev(commands.Cog):
    """Various development focused utilities."""

    async def red_delete_data_for_user(self, **kwargs: Any) -> None:
        """
        Because despite my best efforts to advise otherwise,
        people use ``--dev`` in production
        """
        return

    def __init__(self) -> None:
        super().__init__()
        self._last_result = None
        self.sessions = {}
        self.env_extensions = {}
        self.source_cache = SourceCache()

    def get_environment(self, ctx: commands.Context) -> dict:
        env = {
            "bot": ctx.bot,
            "ctx": ctx,
            "channel": ctx.channel,
            "author": ctx.author,
            "guild": ctx.guild,
            "message": ctx.message,
            "asyncio": asyncio,
            "aiohttp": aiohttp,
            "discord": discord,
            "commands": commands,
            "cf": chat_formatting,
            "_": self._last_result,
            "__name__": "__main__",
        }
        for name, value in self.env_extensions.items():
            try:
                env[name] = value(ctx)
            except Exception as exc:
                traceback.clear_frames(exc.__traceback__)
                env[name] = exc
        return env

    @commands.command()
    @commands.is_owner()
    async def debug(self, ctx: commands.Context, *, code: str) -> None:
        """Evaluate a statement of python code.

        The bot will always respond with the return value of the code.
        If the return value of the code is a coroutine, it will be awaited,
        and the result of that will be the bot's response.

        Note: Only one statement may be evaluated. Using certain restricted
        keywords, e.g. yield, will result in a syntax error. For multiple
        lines or asynchronous code, see [p]repl or [p]eval.

        Environment Variables:
            `ctx`      - the command invocation context
            `bot`      - the bot object
            `channel`  - the current channel object
            `author`   - the command author's member object
            `guild`    - the current guild object
            `message`  - the command's message object
            `_`        - the result of the last dev command
            `aiohttp`  - the aiohttp library
            `asyncio`  - the asyncio library
            `discord`  - the discord.py library
            `commands` - the redbot.core.commands module
            `cf`       - the redbot.core.utils.chat_formatting module
        """
        env = self.get_environment(ctx)
        source = cleanup_code(code)

        output = await DevOutput.from_debug(
            ctx, source=source, source_cache=self.source_cache, env=env
        )
        self._last_result = output.result
        await output.send()

    @commands.command(name="eval")
    @commands.is_owner()
    async def _eval(self, ctx: commands.Context, *, body: str) -> None:
        """Execute asynchronous code.

        This command wraps code into the body of an async function and then
        calls and awaits it. The bot will respond with anything printed to
        stdout, as well as the return value of the function.

        The code can be within a codeblock, inline code or neither, as long
        as they are not mixed and they are formatted correctly.

        Environment Variables:
            `ctx`      - the command invocation context
            `bot`      - the bot object
            `channel`  - the current channel object
            `author`   - the command author's member object
            `guild`    - the current guild object
            `message`  - the command's message object
            `_`        - the result of the last dev command
            `aiohttp`  - the aiohttp library
            `asyncio`  - the asyncio library
            `discord`  - the discord.py library
            `commands` - the redbot.core.commands module
            `cf`       - the redbot.core.utils.chat_formatting module
        """
        env = self.get_environment(ctx)
        source = cleanup_code(body)

        output = await DevOutput.from_eval(
            ctx, source=source, source_cache=self.source_cache, env=env
        )
        if output.result is not None:
            self._last_result = output.result
        await output.send()

    @commands.group(invoke_without_command=True)
    @commands.is_owner()
    async def repl(self, ctx: commands.Context) -> None:
        """Open an interactive REPL.

        The REPL will only recognise code as messages which start with a
        backtick. This includes codeblocks, and as such multiple lines can be
        evaluated.

        Use `exit()` or `quit` to exit the REPL session, prefixed with
        a backtick so they may be interpreted.

        Environment Variables:
            `ctx`      - the command invocation context
            `bot`      - the bot object
            `channel`  - the current channel object
            `author`   - the command author's member object
            `guild`    - the current guild object
            `message`  - the command's message object
            `_`        - the result of the last dev command
            `aiohttp`  - the aiohttp library
            `asyncio`  - the asyncio library
            `discord`  - the discord.py library
            `commands` - the redbot.core.commands module
            `cf`       - the redbot.core.utils.chat_formatting module
        """
        if ctx.channel.id in self.sessions:
            if self.sessions[ctx.channel.id]:
                await ctx.send(
                    _("Already running a REPL session in this channel. Exit it with `quit`.")
                )
            else:
                await ctx.send(
                    _(
                        "Already running a REPL session in this channel. Resume the REPL with `{}repl resume`."
                    ).format(ctx.prefix)
                )
            return

        env = self.get_environment(ctx)
        env["__builtins__"] = __builtins__
        env["_"] = None
        self.sessions[ctx.channel.id] = True
        await ctx.send(
            _(
                "Enter code to execute or evaluate. `exit()` or `quit` to exit. `{}repl pause` to pause."
            ).format(ctx.prefix)
        )

        while True:
            response = await ctx.bot.wait_for("message", check=MessagePredicate.regex(r"^`", ctx))
            env["message"] = response

            if not self.sessions[ctx.channel.id]:
                continue

            source = cleanup_code(response.content)

            if source in ("quit", "exit", "exit()"):
                await ctx.send(_("Exiting."))
                del self.sessions[ctx.channel.id]
                return

            output = await DevOutput.from_repl(
                ctx, source=source, source_cache=self.source_cache, env=env
            )
            try:
                await output.send(tick=False)
            except discord.Forbidden:
                pass
            except discord.HTTPException as exc:
                await ctx.send(_("Unexpected error: ") + str(exc))

    @repl.command(aliases=["resume"])
    async def pause(self, ctx: commands.Context, toggle: Optional[bool] = None) -> None:
        """Pauses/resumes the REPL running in the current channel."""
        if ctx.channel.id not in self.sessions:
            await ctx.send(_("There is no currently running REPL session in this channel."))
            return

        if toggle is None:
            toggle = not self.sessions[ctx.channel.id]
        self.sessions[ctx.channel.id] = toggle

        if toggle:
            await ctx.send(_("The REPL session in this channel has been resumed."))
        else:
            await ctx.send(_("The REPL session in this channel is now paused."))

    @commands.guild_only()
    @commands.command()
    @commands.is_owner()
    async def mock(self, ctx: commands.Context, user: discord.Member, *, command: str) -> None:
        """Mock another user invoking a command.

        The prefix must not be entered.
        """
        msg = copy(ctx.message)
        msg.author = user
        msg.content = ctx.prefix + command

        ctx.bot.dispatch("message", msg)

    @commands.guild_only()
    @commands.command(name="mockmsg")
    @commands.is_owner()
    async def mock_msg(
        self, ctx: commands.Context, user: discord.Member, *, content: str = ""
    ) -> None:
        """Dispatch a message event as if it were sent by a different user.

        Current message is used as a base (including attachments, embeds, etc.),
        the content and author of the message are replaced with the given arguments.

        Note: If `content` isn't passed, the message needs to contain embeds, attachments,
        or anything else that makes the message non-empty.
        """
        msg = ctx.message
        if not content and not msg.embeds and not msg.attachments and not msg.stickers:
            await ctx.send_help()
            return
        msg = copy(msg)
        msg.author = user
        msg.content = content

        ctx.bot.dispatch("message", msg)

    @commands.command()
    @commands.is_owner()
    async def bypasscooldowns(self, ctx: commands.Context, toggle: Optional[bool] = None) -> None:
        """Give bot owners the ability to bypass cooldowns.

        Does not persist through restarts."""
        if toggle is None:
            toggle = not ctx.bot._bypass_cooldowns
        ctx.bot._bypass_cooldowns = toggle

        if toggle:
            await ctx.send(_("Bot owners will now bypass all commands with cooldowns."))
        else:
            await ctx.send(_("Bot owners will no longer bypass all commands with cooldowns."))



from .commands import (
    bot_has_permissions,
    bot_in_a_guild,
    has_permissions,
    is_owner,
    guildowner,
    guildowner_or_permissions,
    admin,
    admin_or_permissions,
    mod,
    mod_or_permissions,
)


__all__ = [
    "bot_has_permissions",
    "bot_in_a_guild",
    "has_permissions",
    "is_owner",
    "guildowner",
    "guildowner_or_permissions",
    "admin",
    "admin_or_permissions",
    "mod",
    "mod_or_permissions",
]



import asyncio
from unittest.mock import patch
import pytest
from collections import Counter


# region Register Tests
async def test_config_register_global(config):
    config.register_global(enabled=False)
    assert config.defaults["GLOBAL"]["enabled"] is False
    assert await config.enabled() is False


def test_config_register_global_badvalues(config):
    with pytest.raises(RuntimeError):
        config.register_global(**{"invalid var name": True})


async def test_config_register_guild(config, empty_guild):
    config.register_guild(enabled=False, some_list=[], some_dict={})
    assert config.defaults[config.GUILD]["enabled"] is False
    assert config.defaults[config.GUILD]["some_list"] == []
    assert config.defaults[config.GUILD]["some_dict"] == {}

    assert await config.guild(empty_guild).enabled() is False
    assert await config.guild(empty_guild).some_list() == []
    assert await config.guild(empty_guild).some_dict() == {}


async def test_config_register_channel(config, empty_channel):
    config.register_channel(enabled=False)
    assert config.defaults[config.CHANNEL]["enabled"] is False
    assert await config.channel(empty_channel).enabled() is False


async def test_config_register_role(config, empty_role):
    config.register_role(enabled=False)
    assert config.defaults[config.ROLE]["enabled"] is False
    assert await config.role(empty_role).enabled() is False


async def test_config_register_member(config, empty_member):
    config.register_member(some_number=-1)
    assert config.defaults[config.MEMBER]["some_number"] == -1
    assert await config.member(empty_member).some_number() == -1


async def test_config_register_user(config, empty_user):
    config.register_user(some_value=None)
    assert config.defaults[config.USER]["some_value"] is None
    assert await config.user(empty_user).some_value() is None


async def test_config_force_register_global(config_fr):
    with pytest.raises(AttributeError):
        await config_fr.enabled()

    config_fr.register_global(enabled=True)
    assert await config_fr.enabled() is True


# endregion


# Test nested registration
async def test_nested_registration(config):
    config.register_global(foo__bar__baz=False)
    assert await config.foo.bar.baz() is False


async def test_nested_registration_asdict(config):
    defaults = {"bar": {"baz": False}}
    config.register_global(foo=defaults)

    assert await config.foo.bar.baz() is False


async def test_nested_registration_and_changing(config):
    defaults = {"bar": {"baz": False}}
    config.register_global(foo=defaults)

    assert await config.foo.bar.baz() is False

    with pytest.raises(ValueError):
        await config.foo.set(True)


async def test_doubleset_default(config):
    config.register_global(foo=True)
    config.register_global(foo=False)

    assert await config.foo() is False


async def test_nested_registration_multidict(config):
    defaults = {"foo": {"bar": {"baz": True}}, "blah": True}
    config.register_global(**defaults)

    assert await config.foo.bar.baz() is True
    assert await config.blah() is True


def test_nested_group_value_badreg(config):
    config.register_global(foo=True)
    with pytest.raises(KeyError):
        config.register_global(foo__bar=False)


async def test_nested_toplevel_reg(config):
    defaults = {"bar": True, "baz": False}
    config.register_global(foo=defaults)

    assert await config.foo.bar() is True
    assert await config.foo.baz() is False


async def test_nested_overlapping(config):
    config.register_global(foo__bar=True)
    config.register_global(foo__baz=False)

    assert await config.foo.bar() is True
    assert await config.foo.baz() is False


async def test_nesting_nofr(config):
    config.register_global(foo={})
    assert await config.foo.bar() is None
    assert await config.foo() == {}


# region Default Value Overrides
async def test_global_default_override(config):
    assert await config.enabled(True) is True


async def test_global_default_nofr(config):
    assert await config.nofr() is None
    assert await config.nofr(True) is True


async def test_guild_default_override(config, empty_guild):
    assert await config.guild(empty_guild).enabled(True) is True


async def test_channel_default_override(config, empty_channel):
    assert await config.channel(empty_channel).enabled(True) is True


async def test_role_default_override(config, empty_role):
    assert await config.role(empty_role).enabled(True) is True


async def test_member_default_override(config, empty_member):
    assert await config.member(empty_member).enabled(True) is True


async def test_user_default_override(config, empty_user):
    assert await config.user(empty_user).some_value(True) is True


# endregion


# region Setting Values
async def test_set_global(config):
    await config.enabled.set(True)
    assert await config.enabled() is True


async def test_set_guild(config, empty_guild):
    await config.guild(empty_guild).enabled.set(True)
    assert await config.guild(empty_guild).enabled() is True

    curr_list = await config.guild(empty_guild).some_list([1, 2, 3])
    assert curr_list == [1, 2, 3]
    curr_list.append(4)

    await config.guild(empty_guild).some_list.set(curr_list)
    assert await config.guild(empty_guild).some_list() == curr_list


async def test_set_channel(config, empty_channel):
    await config.channel(empty_channel).enabled.set(True)
    assert await config.channel(empty_channel).enabled() is True


async def test_set_channel_no_register(config, empty_channel):
    await config.channel(empty_channel).no_register.set(True)
    assert await config.channel(empty_channel).no_register() is True


# endregion


# Dynamic attribute testing
async def test_set_dynamic_attr(config):
    await config.set_raw("foobar", value=True)

    assert await config.foobar() is True


async def test_clear_dynamic_attr(config):
    await config.foo.set(True)
    await config.clear_raw("foo")

    with pytest.raises(KeyError):
        await config.get_raw("foo")


async def test_get_dynamic_attr(config):
    assert await config.get_raw("foobaz", default=True) is True


# Member Group testing
async def test_membergroup_allguilds(config, empty_member):
    await config.member(empty_member).foo.set(False)

    all_servers = await config.all_members()
    assert empty_member.guild.id in all_servers


async def test_membergroup_allmembers(config, empty_member):
    await config.member(empty_member).foo.set(False)

    all_members = await config.all_members(empty_member.guild)
    assert empty_member.id in all_members


# Clearing testing
async def test_global_clear(config):
    config.register_global(foo=True, bar=False)

    await config.foo.set(False)
    await config.bar.set(True)

    assert await config.foo() is False
    assert await config.bar() is True

    await config.clear()

    assert await config.foo() is True
    assert await config.bar() is False


async def test_member_clear(config, member_factory):
    config.register_member(foo=True)

    m1 = member_factory.get()
    await config.member(m1).foo.set(False)
    assert await config.member(m1).foo() is False

    m2 = member_factory.get()
    await config.member(m2).foo.set(False)
    assert await config.member(m2).foo() is False

    assert m1.guild.id != m2.guild.id

    await config.member(m1).clear()
    assert await config.member(m1).foo() is True
    assert await config.member(m2).foo() is False


async def test_member_clear_all(config, member_factory):
    server_ids = []
    for _ in range(5):
        member = member_factory.get()
        await config.member(member).foo.set(True)
        server_ids.append(member.guild.id)

    member = member_factory.get()
    assert len(await config.all_members()) == len(server_ids)

    await config.clear_all_members()

    assert len(await config.all_members()) == 0


async def test_clear_all(config):
    await config.foo.set(True)
    assert await config.foo() is True

    await config.clear_all()
    with pytest.raises(KeyError):
        await config.get_raw("foo")


async def test_clear_value(config):
    await config.foo.set(True)
    await config.foo.clear()

    with pytest.raises(KeyError):
        await config.get_raw("foo")


# Get All testing
async def test_user_get_all_from_kind(config, user_factory):
    config.register_user(foo=False, bar=True)
    for _ in range(5):
        user = user_factory.get()
        await config.user(user).foo.set(True)

    all_data = await config.all_users()

    assert len(all_data) == 5

    for _, v in all_data.items():
        assert v["foo"] is True
        assert v["bar"] is True


async def test_user_getalldata(config, user_factory):
    user = user_factory.get()
    config.register_user(foo=True, bar=False)
    await config.user(user).foo.set(False)

    all_data = await config.user(user).all()

    assert "foo" in all_data
    assert "bar" in all_data

    assert config.user(user).defaults["foo"] is True


async def test_value_ctxmgr(config):
    config.register_global(foo_list=[])

    async with config.foo_list() as foo_list:
        foo_list.append("foo")

    foo_list = await config.foo_list()

    assert "foo" in foo_list


async def test_value_ctxmgr_saves(config):
    config.register_global(bar_list=[])

    try:
        async with config.bar_list() as bar_list:
            bar_list.append("bar")
            raise RuntimeError()
    except RuntimeError:
        pass

    bar_list = await config.bar_list()

    assert "bar" in bar_list


async def test_value_ctxmgr_immutable(config):
    config.register_global(foo=True)

    with pytest.raises(TypeError):
        async with config.foo() as foo:
            foo = False

    foo = await config.foo()
    assert foo is True


async def test_ctxmgr_no_shared_default(config, member_factory):
    config.register_member(foo=[])
    m1 = member_factory.get()
    m2 = member_factory.get()

    async with config.member(m1).foo() as foo:
        foo.append(1)

    assert 1 not in await config.member(m2).foo()


async def test_ctxmgr_no_unnecessary_write(config):
    config.register_global(foo=[])
    foo_value_obj = config.foo
    with patch.object(foo_value_obj, "set") as set_method:
        async with foo_value_obj() as foo:
            pass
        set_method.assert_not_called()


async def test_get_then_mutate(config):
    """Tests that mutating an object after getting it as a value doesn't mutate the data store."""
    config.register_global(list1=[])
    await config.list1.set([])
    list1 = await config.list1()
    list1.append("foo")
    list1 = await config.list1()
    assert "foo" not in list1


async def test_set_then_mutate(config):
    """Tests that mutating an object after setting it as a value doesn't mutate the data store."""
    config.register_global(list1=[])
    list1 = []
    await config.list1.set(list1)
    list1.append("foo")
    list1 = await config.list1()
    assert "foo" not in list1


async def test_call_group_fills_defaults(config):
    config.register_global(subgroup={"foo": True})
    subgroup = await config.subgroup()
    assert "foo" in subgroup


async def test_group_call_ctxmgr_writes(config):
    config.register_global(subgroup={"foo": True})
    async with config.subgroup() as subgroup:
        subgroup["bar"] = False

    subgroup = await config.subgroup()
    assert subgroup == {"foo": True, "bar": False}


async def test_all_works_as_ctxmgr(config):
    config.register_global(subgroup={"foo": True})
    async with config.subgroup.all() as subgroup:
        subgroup["bar"] = False

    subgroup = await config.subgroup()
    assert subgroup == {"foo": True, "bar": False}


async def test_get_raw_mixes_defaults(config):
    config.register_global(subgroup={"foo": True})
    await config.subgroup.set_raw("bar", value=False)

    subgroup = await config.get_raw("subgroup")
    assert subgroup == {"foo": True, "bar": False}


async def test_cast_str_raw(config):
    await config.set_raw(123, 456, value=True)
    assert await config.get_raw(123, 456) is True
    assert await config.get_raw("123", "456") is True
    await config.clear_raw("123", 456)


async def test_cast_str_nested(config):
    config.register_global(foo={})
    await config.foo.set({123: True, 456: {789: False}})
    assert await config.foo() == {"123": True, "456": {"789": False}}


def test_config_custom_noinit(config):
    with pytest.raises(ValueError):
        config.custom("TEST", 1, 2, 3)


def test_config_custom_init(config):
    config.init_custom("TEST", 3)
    config.custom("TEST", 1, 2, 3)


def test_config_custom_doubleinit(config):
    config.init_custom("TEST", 3)
    with pytest.raises(ValueError):
        config.init_custom("TEST", 2)


async def test_config_locks_cache(config, empty_guild):
    lock1 = config.foo.get_lock()
    assert lock1 is config.foo.get_lock()
    lock2 = config.guild(empty_guild).foo.get_lock()
    assert lock2 is config.guild(empty_guild).foo.get_lock()
    assert lock1 is not lock2


async def test_config_value_atomicity(config):
    config.register_global(foo=[])
    tasks = []
    for _ in range(15):

        async def func():
            async with config.foo.get_lock():
                foo = await config.foo()
                foo.append(0)
                await asyncio.sleep(0.1)
                await config.foo.set(foo)

        tasks.append(asyncio.create_task(func()))

    await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)

    assert len(await config.foo()) == 15


async def test_config_ctxmgr_atomicity(config):
    config.register_global(foo=[])
    tasks = []
    for _ in range(15):

        async def func():
            async with config.foo() as foo:
                foo.append(0)
                await asyncio.sleep(0.1)

        tasks.append(asyncio.create_task(func()))

    await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)

    assert len(await config.foo()) == 15


async def test_set_with_partial_primary_keys(config):
    config.init_custom("CUSTOM", 3)
    await config.custom("CUSTOM", "1").set({"11": {"111": {"foo": "bar"}}})
    assert await config.custom("CUSTOM", "1", "11", "111").foo() == "bar"

    await config.custom("CUSTOM", "2").set(
        {
            "11": {"111": {"foo": "bad"}},
            "22": {"111": {"foo": "baz"}},
            "33": {"111": {"foo": "boo"}, "222": {"foo": "boz"}},
        }
    )
    assert await config.custom("CUSTOM", "2", "11", "111").foo() == "bad"
    assert await config.custom("CUSTOM", "2", "22", "111").foo() == "baz"
    assert await config.custom("CUSTOM", "2", "33", "111").foo() == "boo"
    assert await config.custom("CUSTOM", "2", "33", "222").foo() == "boz"

    await config.custom("CUSTOM", "2").set({"22": {}, "33": {"111": {}, "222": {"foo": "biz"}}})
    with pytest.raises(KeyError):
        await config.custom("CUSTOM").get_raw("2", "11")
    with pytest.raises(KeyError):
        await config.custom("CUSTOM").get_raw("2", "22", "111")
    with pytest.raises(KeyError):
        await config.custom("CUSTOM").get_raw("2", "33", "111", "foo")
    assert await config.custom("CUSTOM", "2", "33", "222").foo() == "biz"


async def test_raw_with_partial_primary_keys(config):
    config.init_custom("CUSTOM", 1)
    await config.custom("CUSTOM").set_raw("primary_key", "identifier", value=True)
    assert await config.custom("CUSTOM", "primary_key").identifier() is True
    await config.custom("CUSTOM").set_raw(value={"primary_key": {"identifier": False}})
    assert await config.custom("CUSTOM", "primary_key").identifier() is False


@pytest.mark.asyncio
async def test_cast_subclass_default(config):
    # regression test for GH-5557/GH-5585
    config.register_global(foo=Counter({}))
    assert type(config.defaults["GLOBAL"]["foo"]) is dict
    assert config.defaults["GLOBAL"]["foo"] == {}
    stored_value = await config.foo()
    assert type(stored_value) is dict
    assert stored_value == {}
    await config.foo.set(Counter({"bar": 1}))
    stored_value = await config.foo()
    assert type(stored_value) is dict
    assert stored_value == {"bar": 1}


"""
Following PARAMS can be generated with:
from functools import reduce
from pprint import pprint
def generate_test_args(print_args=True):
    pkeys = ("1", "2", "3")
    identifiers = ("foo",)
    full_dict = {"1": {"2": {"3": {"foo": "bar"}}}}
    argvalues = [
        (
            pkeys[:x],
            (pkeys[x:] + identifiers)[:y],
            reduce(lambda d, k: d[k], (pkeys + identifiers)[:x+y], full_dict),
        )
        for x in range(len(pkeys) + 1)
        for y in range(len(pkeys) + len(identifiers) - x + 1)
    ]
    if print_args:
        print("[")
        for args in argvalues:
            print(f"    {args!r},")
        print("]")
    else:
        return argvalues
generate_test_args()
"""
PARAMS = [
    ((), (), {"1": {"2": {"3": {"foo": "bar"}}}}),
    ((), (1,), {"2": {"3": {"foo": "bar"}}}),
    ((), (1, 2), {"3": {"foo": "bar"}}),
    ((), (1, 2, 3), {"foo": "bar"}),
    ((), (1, 2, 3, "foo"), "bar"),
    ((1,), (), {"2": {"3": {"foo": "bar"}}}),
    ((1,), (2,), {"3": {"foo": "bar"}}),
    ((1,), (2, 3), {"foo": "bar"}),
    ((1,), (2, 3, "foo"), "bar"),
    ((1, 2), (), {"3": {"foo": "bar"}}),
    ((1, 2), (3,), {"foo": "bar"}),
    ((1, 2), (3, "foo"), "bar"),
    ((1, 2, 3), (), {"foo": "bar"}),
    ((1, 2, 3), ("foo",), "bar"),
]


@pytest.mark.parametrize("pkeys, raw_args, result", PARAMS)
async def test_config_custom_partial_pkeys_get(config, pkeys, raw_args, result):
    # setup
    config.init_custom("TEST", 3)
    config.register_custom("TEST")
    await config.custom("TEST", 1, 2, 3).set({"foo": "bar"})

    group = config.custom("TEST", *pkeys)
    assert await group.get_raw(*raw_args) == result


@pytest.mark.parametrize("pkeys, raw_args, result", PARAMS)
async def test_config_custom_partial_pkeys_set(config, pkeys, raw_args, result):
    # setup
    config.init_custom("TEST", 3)
    config.register_custom("TEST")
    await config.custom("TEST", 1, 2, 3).set({"foo": "blah"})

    group = config.custom("TEST", *pkeys)
    await group.set_raw(*raw_args, value=result)
    assert await group.get_raw(*raw_args) == result


@pytest.mark.asyncio
async def test_config_custom_get_raw_with_default_on_whole_scope(config):
    config.init_custom("TEST", 3)
    config.register_custom("TEST")

    group = config.custom("TEST")
    assert await group.get_raw(default=True) is True


@pytest.mark.parametrize(
    "pkeys,raw_args,to_set",
    (
        # no config data for (cog_name, cog_id) is present
        ((), (), None),
        ((1,), (), None),
        ((1, 2), (), None),
        ((1, 2, 3), (), None),
        ((1, 2, 3), ("key1",), None),
        ((1, 2, 3), ("key1", "key2"), None),
        # config data for (cog_name, cog_id) is present but scope does not exist
        ((), (), ()),
        ((1,), (), ()),
        ((1, 2), (), ()),
        ((1, 2, 3), (), ()),
        ((1, 2, 3), ("key1",), ()),
        ((1, 2, 3), ("key1", "key2"), ()),
        # the scope exists with no records
        ((1,), (), ("1",)),
        ((1, 2), (), ("1",)),
        ((1, 2, 3), (), ("1",)),
        ((1, 2, 3), ("key1",), ("1",)),
        ((1, 2, 3), ("key1", "key2"), ("1",)),
        # scope with partial primary key (1,) exists
        ((1, 2), (), ("1", "2")),
        ((1, 2, 3), (), ("1", "2")),
        ((1, 2, 3), ("key1",), ("1", "2")),
        ((1, 2, 3), ("key1", "key2"), ("1", "2")),
        # scope with partial primary key (1, 2) exists
        ((1, 2, 3), (), ("1", "2", "3")),
        ((1, 2, 3), ("key1",), ("1", "2", "3")),
        ((1, 2, 3), ("key1", "key2"), ("1", "2", "3")),
        # scope with full primary key (1, 2, 3)
        ((1, 2, 3), ("key1",), ("1", "2", "3", "key1")),
        ((1, 2, 3), ("key1", "key2"), ("1", "2", "3", "key1")),
        # scope with full primary key (1, 2, 3) and a group named "key1" exists
        ((1, 2, 3), ("key1", "key2"), ("1", "2", "3", "key1", "key2")),
    ),
)
@pytest.mark.asyncio
async def test_config_custom_clear_identifiers_that_do_not_exist(config, pkeys, raw_args, to_set):
    config.init_custom("TEST", 3)
    config.register_custom("TEST")

    group = config.custom("TEST", *pkeys)
    if to_set is not None:
        data = {}
        partial = data
        for key in to_set:
            partial[key] = {}
            partial = partial[key]
        scope = config.custom("TEST")
        await scope.set(data)
        # Clear needed to be able to differ between missing config data and missing scope data
        await scope.clear_raw(*to_set)
    await group.clear_raw(*raw_args)



import importlib.metadata
import os
import sys
from packaging.requirements import Requirement
from packaging.version import Version

import pytest

from redbot import core
from redbot.core import VersionInfo


def test_version_working():
    assert hasattr(core, "__version__")
    assert core.__version__[0] == "3"


# When adding more of these, ensure they are added in ascending order of precedence
version_tests = (
    "3.0.0.dev1",
    "3.0.0.dev2",
    "3.0.0a32.dev12",
    "3.0.0a32",
    "3.0.0a32.post10.dev12",
    "3.0.0a32.post10",
    "3.0.0b23.dev4",
    "3.0.0b23",
    "3.0.0b23.post5.dev16",
    "3.0.0b23.post5",
    "3.0.0rc1.dev1",
    "3.0.0rc1",
    "3.0.0",
    "3.0.0.post1.dev1",
    "3.0.1.dev1",
    "3.0.1.dev2+gdbaf31e",
    "3.0.1.dev2+gdbaf31e.dirty",
    "3.0.1.dev3+gae98d77",
    "3.0.1",
    "3.0.1.post1.dev1",
    "3.0.1.post1",
    "2018.10.6b21",
)


def test_version_info_str_parsing():
    for version_str in version_tests:
        assert version_str == str(VersionInfo.from_str(version_str))


def test_version_info_lt():
    for version_cls in (Version, VersionInfo.from_str):
        for next_idx, cur in enumerate(version_tests[:-1], start=1):
            cur_test = version_cls(cur)
            next_test = version_cls(version_tests[next_idx])
            assert cur_test < next_test


def test_version_info_gt():
    assert VersionInfo.from_str(version_tests[1]) > VersionInfo.from_str(version_tests[0])


def test_python_version_has_lower_bound():
    """
    Due to constant issues in support with Red being installed on a Python version that was not
    supported by any Red version, it is important that we have both an upper and lower bound set.
    """
    requires_python = importlib.metadata.metadata("Red-DiscordBot").get("Requires-Python")
    assert requires_python is not None

    # Requirement needs a regular requirement string, so "x" serves as requirement's name here
    req = Requirement(f"x{requires_python}")
    assert any(spec.operator in (">", ">=") for spec in req.specifier)


@pytest.mark.skipif(
    os.getenv("TOX_RED", False) and sys.version_info >= (3, 12),
    reason="Testing on yet to be supported Python version.",
)
def test_python_version_has_upper_bound():
    """
    Due to constant issues in support with Red being installed on a Python version that was not
    supported by any Red version, it is important that we have both an upper and lower bound set.
    """
    requires_python = importlib.metadata.metadata("Red-DiscordBot").get("Requires-Python")
    assert requires_python is not None

    # Requirement needs a regular requirement string, so "x" serves as requirement's name here
    req = Requirement(f"x{requires_python}")
    assert any(spec.operator in ("<", "<=") for spec in req.specifier)



from __future__ import annotations

import datetime
import itertools
import math
import textwrap
from io import BytesIO, StringIO
from typing import Any, Iterator, List, Literal, Optional, Sequence, SupportsInt, Union

import discord
from babel.lists import format_list as babel_list
from babel.numbers import format_decimal
from rich.console import Console

from redbot.core.i18n import Translator, get_babel_locale, get_babel_regional_format

__all__ = (
    "error",
    "warning",
    "info",
    "success",
    "question",
    "bold",
    "box",
    "header",
    "hyperlink",
    "inline",
    "italics",
    "spoiler",
    "pagify",
    "strikethrough",
    "subtext",
    "underline",
    "quote",
    "escape",
    "humanize_list",
    "format_perms_list",
    "humanize_timedelta",
    "humanize_number",
    "text_to_file",
    "rich_markup",
)

_ = Translator("UtilsChatFormatting", __file__)


def hyperlink(text: str, url: str) -> str:
    """Create hyperlink markdown with text and a URL.

    Parameters
    ----------
    text : str
        The text which will contain the link.
    url : str
        The URL used for the hyperlink.

    Returns
    -------
    str
        The new message.

    """
    return f"[{text}]({url})"


def header(text: str, size: Literal["small", "medium", "large"]) -> str:
    """Formats a header.

    Parameters
    ----------
    text : str
        The text for the header.
    size : Literal['small', 'medium', 'large']
        The size of the header ('small', 'medium' or 'large')

    Returns
    -------
    str
        The new message.

    """
    if size == "small":
        multiplier = 3
    elif size == "medium":
        multiplier = 2
    elif size == "large":
        multiplier = 1
    else:
        raise ValueError(f"Invalid size '{size}'")
    return "#" * multiplier + " " + text


def subtext(text: str) -> str:
    """Formats subtext from the given text.

    Parameters
    ----------
    text : str
        The text to format as subtext.

    Returns
    -------
    str
        The new message.

    """
    return "-# " + text


def error(text: str) -> str:
    """Get text prefixed with an error emoji.

    Parameters
    ----------
    text : str
        The text to be prefixed.

    Returns
    -------
    str
        The new message.

    """
    return f"\N{NO ENTRY SIGN} {text}"


def warning(text: str) -> str:
    """Get text prefixed with a warning emoji.

    Parameters
    ----------
    text : str
        The text to be prefixed.

    Returns
    -------
    str
        The new message.

    """
    return f"\N{WARNING SIGN}\N{VARIATION SELECTOR-16} {text}"


def info(text: str) -> str:
    """Get text prefixed with an info emoji.

    Parameters
    ----------
    text : str
        The text to be prefixed.

    Returns
    -------
    str
        The new message.

    """
    return f"\N{INFORMATION SOURCE}\N{VARIATION SELECTOR-16} {text}"


def success(text: str) -> str:
    """Get text prefixed with a success emoji.

    Parameters
    ----------
    text : str
        The text to be prefixed.

    Returns
    -------
    str
        The new message.

    """
    return f"\N{WHITE HEAVY CHECK MARK} {text}"


def question(text: str) -> str:
    """Get text prefixed with a question emoji.

    Parameters
    ----------
    text : str
        The text to be prefixed.

    Returns
    -------
    str
        The new message.

    """
    return f"\N{BLACK QUESTION MARK ORNAMENT}\N{VARIATION SELECTOR-16} {text}"


def bold(text: str, escape_formatting: bool = True) -> str:
    """Get the given text in bold.

    Note: By default, this function will escape ``text`` prior to emboldening.

    Parameters
    ----------
    text : str
        The text to be marked up.
    escape_formatting : `bool`, optional
        Set to :code:`False` to not escape markdown formatting in the text.

    Returns
    -------
    str
        The marked up text.

    """
    return f"**{escape(text, formatting=escape_formatting)}**"


def box(text: str, lang: str = "") -> str:
    """Get the given text in a code block.

    Parameters
    ----------
    text : str
        The text to be marked up.
    lang : `str`, optional
        The syntax highlighting language for the codeblock.

    Returns
    -------
    str
        The marked up text.

    """
    return f"```{lang}\n{text}\n```"


def inline(text: str) -> str:
    """Get the given text as inline code.

    Parameters
    ----------
    text : str
        The text to be marked up.

    Returns
    -------
    str
        The marked up text.

    """
    if "`" in text:
        return f"``{text}``"
    else:
        return f"`{text}`"


def italics(text: str, escape_formatting: bool = True) -> str:
    """Get the given text in italics.

    Note: By default, this function will escape ``text`` prior to italicising.

    Parameters
    ----------
    text : str
        The text to be marked up.
    escape_formatting : `bool`, optional
        Set to :code:`False` to not escape markdown formatting in the text.

    Returns
    -------
    str
        The marked up text.

    """
    return f"*{escape(text, formatting=escape_formatting)}*"


def spoiler(text: str, escape_formatting: bool = True) -> str:
    """Get the given text as a spoiler.

    Note: By default, this function will escape ``text`` prior to making the text a spoiler.

    Parameters
    ----------
    text : str
        The text to be marked up.
    escape_formatting : `bool`, optional
        Set to :code:`False` to not escape markdown formatting in the text.

    Returns
    -------
    str
        The marked up text.

    """
    return f"||{escape(text, formatting=escape_formatting)}||"


class pagify(Iterator[str]):
    """Generate multiple pages from the given text.

    The returned iterator supports length estimation with :func:`operator.length_hint()`.

    Note
    ----
    This does not respect code blocks or inline code.

    Parameters
    ----------
    text : str
        The content to pagify and send.
    delims : `sequence` of `str`, optional
        Characters where page breaks will occur. If no delimiters are found
        in a page, the page will break after ``page_length`` characters.
        By default this only contains the newline.

    Other Parameters
    ----------------
    priority : `bool`
        Set to :code:`True` to choose the page break delimiter based on the
        order of ``delims``. Otherwise, the page will always break at the
        last possible delimiter.
    escape_mass_mentions : `bool`
        If :code:`True`, any mass mentions (here or everyone) will be
        silenced.
    shorten_by : `int`
        How much to shorten each page by. Defaults to 8.
    page_length : `int`
        The maximum length of each page. Defaults to 2000.

    Yields
    ------
    `str`
        Pages of the given text.

    """

    # when changing signature of this method, please update it in docs/framework_utils.rst as well
    def __init__(
        self,
        text: str,
        delims: Sequence[str] = ("\n",),
        *,
        priority: bool = False,
        escape_mass_mentions: bool = True,
        shorten_by: int = 8,
        page_length: int = 2000,
    ) -> None:
        self._text = text
        self._delims = delims
        self._priority = priority
        self._escape_mass_mentions = escape_mass_mentions
        self._shorten_by = shorten_by
        self._page_length = page_length - shorten_by

        self._start = 0
        self._end = len(text)

    def __repr__(self) -> str:
        text = self._text
        if len(text) > 20:
            text = f"{text[:19]}\N{HORIZONTAL ELLIPSIS}"
        return (
            "pagify("
            f"{text!r},"
            f" {self._delims!r},"
            f" priority={self._priority!r},"
            f" escape_mass_mentions={self._escape_mass_mentions!r},"
            f" shorten_by={self._shorten_by!r},"
            f" page_length={self._page_length + self._shorten_by!r}"
            ")"
        )

    def __length_hint__(self) -> int:
        return math.ceil((self._end - self._start) / self._page_length)

    def __iter__(self) -> pagify:
        return self

    def __next__(self) -> str:
        text = self._text
        escape_mass_mentions = self._escape_mass_mentions
        page_length = self._page_length
        start = self._start
        end = self._end

        while (end - start) > page_length:
            stop = start + page_length
            if escape_mass_mentions:
                stop -= text.count("@here", start, stop) + text.count("@everyone", start, stop)
            closest_delim_it = (text.rfind(d, start + 1, stop) for d in self._delims)
            if self._priority:
                closest_delim = next((x for x in closest_delim_it if x > 0), -1)
            else:
                closest_delim = max(closest_delim_it)
            stop = closest_delim if closest_delim != -1 else stop
            if escape_mass_mentions:
                to_send = escape(text[start:stop], mass_mentions=True)
            else:
                to_send = text[start:stop]
            start = self._start = stop
            if len(to_send.strip()) > 0:
                return to_send

        if len(text[start:end].strip()) > 0:
            self._start = end
            if escape_mass_mentions:
                return escape(text[start:end], mass_mentions=True)
            else:
                return text[start:end]

        raise StopIteration


def strikethrough(text: str, escape_formatting: bool = True) -> str:
    """Get the given text with a strikethrough.

    Note: By default, this function will escape ``text`` prior to applying a strikethrough.

    Parameters
    ----------
    text : str
        The text to be marked up.
    escape_formatting : `bool`, optional
        Set to :code:`False` to not escape markdown formatting in the text.

    Returns
    -------
    str
        The marked up text.

    """
    return f"~~{escape(text, formatting=escape_formatting)}~~"


def underline(text: str, escape_formatting: bool = True) -> str:
    """Get the given text with an underline.

    Note: By default, this function will escape ``text`` prior to underlining.

    Parameters
    ----------
    text : str
        The text to be marked up.
    escape_formatting : `bool`, optional
        Set to :code:`False` to not escape markdown formatting in the text.

    Returns
    -------
    str
        The marked up text.

    """
    return f"__{escape(text, formatting=escape_formatting)}__"


def quote(text: str) -> str:
    """Quotes the given text.

    Parameters
    ----------
    text : str
        The text to be marked up.

    Returns
    -------
    str
        The marked up text.

    """
    return textwrap.indent(text, "> ", lambda l: True)


def escape(text: str, *, mass_mentions: bool = False, formatting: bool = False) -> str:
    """Get text with all mass mentions or markdown escaped.

    Parameters
    ----------
    text : str
        The text to be escaped.
    mass_mentions : `bool`, optional
        Set to :code:`True` to escape mass mentions in the text.
    formatting : `bool`, optional
        Set to :code:`True` to escape any markdown formatting in the text.

    Returns
    -------
    str
        The escaped text.

    """
    if mass_mentions:
        text = text.replace("@everyone", "@\u200beveryone")
        text = text.replace("@here", "@\u200bhere")
    if formatting:
        text = discord.utils.escape_markdown(text)
    return text


def humanize_list(
    items: Sequence[str], *, locale: Optional[str] = None, style: str = "standard"
) -> str:
    """Get comma-separated list, with the last element joined with *and*.

    Parameters
    ----------
    items : Sequence[str]
        The items of the list to join together.
    locale : Optional[str]
        The locale to convert, if not specified it defaults to the bot's locale.
    style : str
        The style to format the list with.

        Note: Not all styles are necessarily available in all locales,
        see documentation of `babel.lists.format_list` for more details.

        standard
            A typical 'and' list for arbitrary placeholders.
            eg. "January, February, and March"
        standard-short
             A short version of a 'and' list, suitable for use with short or
             abbreviated placeholder values.
             eg. "Jan., Feb., and Mar."
        or
            A typical 'or' list for arbitrary placeholders.
            eg. "January, February, or March"
        or-short
            A short version of an 'or' list.
            eg. "Jan., Feb., or Mar."
        unit
            A list suitable for wide units.
            eg. "3 feet, 7 inches"
        unit-short
            A list suitable for short units
            eg. "3 ft, 7 in"
        unit-narrow
            A list suitable for narrow units, where space on the screen is very limited.
            eg. "3′ 7″"

    Raises
    ------
    ValueError
        The locale does not support the specified style.

    Examples
    --------
    .. testsetup::

        from redbot.core.utils.chat_formatting import humanize_list

    .. doctest::

        >>> humanize_list(['One', 'Two', 'Three'])
        'One, Two, and Three'
        >>> humanize_list(['One'])
        'One'
        >>> humanize_list(['omena', 'peruna', 'aplari'], style='or', locale='fi')
        'omena, peruna tai aplari'

    """

    return babel_list(items, style=style, locale=get_babel_locale(locale))


def format_perms_list(perms: discord.Permissions) -> str:
    """Format a list of permission names.

    This will return a humanized list of the names of all enabled
    permissions in the provided `discord.Permissions` object.

    Parameters
    ----------
    perms : discord.Permissions
        The permissions object with the requested permissions to list
        enabled.

    Returns
    -------
    str
        The humanized list.

    """
    perm_names: List[str] = []
    for perm, value in perms:
        if value is True:
            perm_name = '"' + perm.replace("_", " ").title() + '"'
            perm_names.append(perm_name)
    return humanize_list(perm_names).replace("Guild", "Server")


def humanize_timedelta(
    *,
    timedelta: Optional[datetime.timedelta] = None,
    seconds: Optional[SupportsInt] = None,
    negative_format: Optional[str] = None,
    maximum_units: Optional[int] = None,
) -> str:
    """
    Get a locale aware human timedelta representation.

    This works with either a timedelta object or a number of seconds.

    Fractional values will be omitted.

    Values that are less than 1 second but greater than -1 second
    will be an empty string.

    Parameters
    ----------
    timedelta: Optional[datetime.timedelta]
        A timedelta object
    seconds: Optional[SupportsInt]
        A number of seconds
    negative_format: Optional[str]
        How to format negative timedeltas, using %-formatting rules.
        Defaults to "negative %s"
    maximum_units: Optional[int]
        The maximum number of different units to output in the final string.

    Returns
    -------
    str
        A locale aware representation of the timedelta or seconds.

    Raises
    ------
    ValueError
        The function was called with neither a number of seconds nor a timedelta object,
        or with a maximum_units less than 1.

    Examples
    --------
    .. testsetup::

        from datetime import timedelta
        from redbot.core.utils.chat_formatting import humanize_timedelta

    .. doctest::

        >>> humanize_timedelta(seconds=314)
        '5 minutes, 14 seconds'
        >>> humanize_timedelta(timedelta=timedelta(minutes=3.14), maximum_units=1)
        '3 minutes'
        >>> humanize_timedelta(timedelta=timedelta(days=-3.14), negative_format="%s ago", maximum_units=3)
        '3 days, 3 hours, 21 minutes ago'
    """

    try:
        obj = seconds if seconds is not None else timedelta.total_seconds()
    except AttributeError:
        raise ValueError("You must provide either a timedelta or a number of seconds")
    if maximum_units is not None and maximum_units < 1:
        raise ValueError("maximum_units must be >= 1")

    periods = [
        (_("year"), _("years"), 60 * 60 * 24 * 365),
        (_("month"), _("months"), 60 * 60 * 24 * 30),
        (_("day"), _("days"), 60 * 60 * 24),
        (_("hour"), _("hours"), 60 * 60),
        (_("minute"), _("minutes"), 60),
        (_("second"), _("seconds"), 1),
    ]
    seconds = int(obj)
    if seconds < 0:
        seconds = -seconds
        if negative_format and "%s" not in negative_format:
            negative_format = negative_format + " %s"
        else:
            negative_format = negative_format or (_("negative") + " %s")
    else:
        negative_format = "%s"
    strings = []
    maximum_units = maximum_units or len(periods)
    for period_name, plural_period_name, period_seconds in periods:
        if seconds >= period_seconds:
            period_value, seconds = divmod(seconds, period_seconds)
            if period_value == 0:
                continue
            unit = plural_period_name if period_value > 1 else period_name
            strings.append(f"{period_value} {unit}")
            if len(strings) == maximum_units:
                break

    return negative_format % humanize_list(strings, style="unit")


def humanize_number(val: Union[int, float], override_locale=None) -> str:
    """
    Convert an int or float to a str with digit separators based on bot locale.

    Parameters
    ----------
    val : Union[int, float]
        The int/float to be formatted.
    override_locale: Optional[str]
        A value to override bot's regional format.

    Raises
    ------
    decimals.InvalidOperation
        If val is greater than 10 x 10^21 for some locales, 10 x 10^24 in others.

    Returns
    -------
    str
        Locale-aware formatted number.
    """
    return format_decimal(val, locale=get_babel_regional_format(override_locale))


def text_to_file(
    text: str, filename: str = "file.txt", *, spoiler: bool = False, encoding: str = "utf-8"
):
    """Prepares text to be sent as a file on Discord, without character limit.

    This writes text into a bytes object that can be used for the ``file`` or ``files`` parameters
    of :meth:`discord.abc.Messageable.send`.

    Parameters
    ----------
    text: str
        The text to put in your file.
    filename: str
        The name of the file sent. Defaults to ``file.txt``.
    spoiler: bool
        Whether the attachment is a spoiler. Defaults to ``False``.

    Returns
    -------
    discord.File
        The file containing your text.

    """
    file = BytesIO(text.encode(encoding))
    return discord.File(file, filename, spoiler=spoiler)


def rich_markup(
    *objects: Any,
    crop: Optional[bool] = True,
    emoji: Optional[bool] = True,
    highlight: Optional[bool] = True,
    justify: Optional[str] = None,
    markup: Optional[bool] = True,
    no_wrap: Optional[bool] = None,
    overflow: Optional[str] = None,
    width: Optional[int] = None,
) -> str:
    """Returns a codeblock with ANSI formatting for colour support.

    This supports a limited set of Rich markup, and rich helper functions. (https://rich.readthedocs.io/en/stable/index.html)

    Parameters
    ----------
    *objects: Any
        The text to convert to ANSI formatting.
    crop: Optional[bool]
        Crop output to width of virtual terminal. Defaults to ``True``.
    emoji: Optional[bool]
        Enable emoji code. Defaults to ``True``.
    highlight: Optional[bool]
        Enable automated highlighting. Defaults to ``True``.
    justify: Optional[str]
        Justify method: "default", "left", "right", "center", or "full". Defaults to ``None``.
    markup: Optional[bool]
        Boolean to enable Console Markup. Defaults to ``True``.
    no_wrap: Optional[bool]
        Disables word wrapping. Defaults to ``None``.
    overflow: Optional[str]
        Overflow method: "ignore", "crop", "fold", or "ellipsis". Defaults to None.
    width: Optional[int]
        The width of the virtual terminal. Defaults to ``80`` characters long.


    Returns
    -------
    str:
        The ANSI formatted text in a codeblock.
    """
    temp_console = Console(  # Prevent messing with STDOUT's console
        color_system="standard",  # Discord only supports 8-bit in colors
        emoji=emoji,
        file=StringIO(),
        force_terminal=True,
        force_interactive=False,
        highlight=highlight,
        markup=markup,
        width=width if width is not None else 80,
    )

    temp_console.print(
        *objects,
        crop=crop,
        justify=justify,
        no_wrap=no_wrap,
        overflow=overflow,
    )
    return box(temp_console.file.getvalue(), lang="ansi")
